import numpy as np
import scipy.sparse as sp
import warnings
import properties

from .base import BaseRegularization, BaseComboRegularization
from .. import Utils


class BaseSparse(BaseRegularization):
    """
    Base class for building up the components of the Sparse Regularization
    """
    def __init__(self, mesh, **kwargs):
        self._stashedR = None
        super(BaseSparse, self).__init__(mesh=mesh, **kwargs)

    model = properties.Array(
        "current model", dtype=float
    )

    epsilon = properties.Float(
        "Threshold value for the model norm", default=1e-3,
        required=True
    )

    norm = properties.Array(
        "norm used", dtype=float
    )

    space = properties.String(
        "By default inherit the objctive", default='linear'
    )

    gradientType = properties.String(
        "type of gradient", default='components'
    )

    scale = properties.Array(
        "General nob for scaling", dtype=float,
    )

    # Give the option to scale or not
    scaledIRLS = properties.Bool(
        "Scale the gradients of the IRLS norms",
        default=True
    )

    @properties.validator('scale')
    def _validate_scale(self, change):
        if change['value'] is not None:
            # todo: residual size? we need to know the expected end shape
            if self._nC_residual != '*':
                assert len(change['value']) == self._nC_residual, (
                    'scale must be length {} not {}'.format(
                        self._nC_residual, len(change['value'])
                    )
                )

    @property
    def stashedR(self):
        return self._stashedR

    @stashedR.setter
    def stashedR(self, value):
        self._stashedR = value


class SparseSmall(BaseSparse):
    """
    Sparse smallness regularization

    **Inputs**

    :param int norm: norm on the smallness
    """

    _multiplier_pair = 'alpha_s'

    def __init__(self, mesh, **kwargs):
        super(SparseSmall, self).__init__(
            mesh=mesh, **kwargs
        )

    # Give the option to scale or not
    scaledIRLS = properties.Bool(
        "Scale the gradients of the IRLS norms",
        default=True
    )

    @property
    def f_m(self):

        return self.mapping * self._delta_m(self.model)

    @property
    def W(self):
        if getattr(self, 'model', None) is None:
            R = Utils.speye(self.mapping.shape[0])
        else:
            r = self.R(self.f_m)
            R = Utils.sdiag(r)

        if self.scale is None:
            self.scale = np.ones(self.mapping.shape[0])

        if self.cell_weights is not None:
            return Utils.sdiag((self.scale *
                                self.cell_weights)**0.5) * R

        else:
            return Utils.sdiag((self.scale * self.regmesh.vol)**0.5) * R

    def R(self, f_m):
        # if R is stashed, return that instead
        if getattr(self, 'stashedR') is not None:
            return self.stashedR

        # Default
        eta = np.ones_like(f_m)

        if self.scaledIRLS:
            # Eta scaling is important for mix-norms...do not mess with it
            # Scale on l2-norm gradient: f_m.max()
            maxVal = np.ones_like(f_m) * np.abs(f_m).max()

            # Compute theoritical maximum gradients for p < 1
            maxVal[self.norm < 1] = (
                self.epsilon / np.sqrt(1.-self.norm[self.norm < 1])
            )
            maxGrad = (
                maxVal /
                (maxVal**2. + self.epsilon**2.)**(1.-self.norm/2.)
            )
            # Scaling factor
            eta[maxGrad != 0] = np.abs(f_m).max()/maxGrad[maxGrad != 0]

        # Scaled IRLS weights
        r = (eta / (f_m**2. + self.epsilon**2.)**(1.-self.norm/2.))**0.5

        self.stashedR = r  # stash on the first calculation
        return r

    @Utils.timeIt
    def deriv(self, m):
        """

        The regularization is:

        .. math::

            R(m) = \\frac{1}{2}\mathbf{(m-m_\\text{ref})^\\top W^\\top
                   W(m-m_\\text{ref})}

        So the derivative is straight forward:

        .. math::

            R(m) = \mathbf{W^\\top W (m-m_\\text{ref})}

        """

        mD = self.mapping.deriv(self._delta_m(m))
        r = self.W * (self.mapping * (self._delta_m(m)))
        return mD.T * (self.W.T * r)


class SparseDeriv(BaseSparse):
    """
    Base Class for sparse regularization on first spatial derivatives
    """

    def __init__(self, mesh, orientation='x', **kwargs):

        self.orientation = orientation
        super(SparseDeriv, self).__init__(mesh=mesh, **kwargs)

    mrefInSmooth = properties.Bool(
        "include mref in the smoothness calculation?", default=False
    )

    # Give the option to scale or not
    scaledIRLS = properties.Bool(
        "Scale the gradients of the IRLS norms",
        default=True
    )

    @Utils.timeIt
    def __call__(self, m):
        """
        We use a weighted 2-norm objective function

        .. math::

            r(m) = \\frac{1}{2}
        """
        if self.mrefInSmooth:

            f_m = self._delta_m(m)

        else:
            f_m = m
        if self.scale is None:
            self.scale = np.ones(self.mapping.shape[0])

        if self.space == 'spherical':
            Ave = getattr(self.regmesh, 'aveCC2F{}'.format(self.orientation))

            if getattr(self, 'model', None) is None:
                R = Utils.speye(self.cellDiffStencil.shape[0])

            else:
                r = self.R(self.f_m)
                R = Utils.sdiag(r)

            if self.cell_weights is not None:
                W = (
                    Utils.sdiag(
                        (Ave*(self.scale * self.cell_weights))**0.5
                    ) *
                    R
                )

            else:
                W = Utils.sdiag(
                    (Ave * (self.scale * self.regmesh.vol))**0.5
                ) * R

            theta = self.cellDiffStencil * (self.mapping * f_m)
            dmdx = Utils.matutils.coterminal(theta)
            r = W * dmdx

        else:
            r = self.W * (self.mapping * f_m)

        return 0.5 * r.dot(r)

    def R(self, f_m):
        # if R is stashed, return that instead
        if getattr(self, 'stashedR') is not None:
            return self.stashedR

        # Default
        eta = np.ones_like(f_m)

        if self.scaledIRLS:
            # Eta scaling is important for mix-norms...do not mess with it
            # Scale on l2-norm gradient: f_m.max()
            maxVal = np.ones_like(f_m) * np.abs(f_m).max()

            # Compute theoritical maximum gradients for p < 1
            maxVal[self.norm < 1] = (
                self.epsilon / np.sqrt(1.-self.norm[self.norm < 1])
            )
            maxGrad = (
                maxVal /
                (maxVal**2. + self.epsilon**2.)**(1.-self.norm/2.)
            )

            # Scaling Factor
            eta[maxGrad != 0] = np.abs(f_m).max()/maxGrad[maxGrad != 0]

        # Scaled-IRLS weights
        r = (eta / (f_m**2. + self.epsilon**2.)**(1.-self.norm/2.))**0.5
        self.stashedR = r  # stash on the first calculation
        return r

    @Utils.timeIt
    def deriv(self, m):
        """

        The regularization is:

        .. math::

            R(m) = \\frac{1}{2}\mathbf{(m-m_\\text{ref})^\\top W^\\top
                   W(m-m_\\text{ref})}

        So the derivative is straight forward:

        .. math::

            R(m) = \mathbf{W^\\top W (m-m_\\text{ref})}

        """

        if self.mrefInSmooth:

            model = self._delta_m(m)

        else:
            model = m
        if self.scale is None:
            self.scale = np.ones(self.mapping.shape[0])

        if self.space == 'spherical':
            Ave = getattr(self.regmesh, 'aveCC2F{}'.format(self.orientation))

            if getattr(self, 'model', None) is None:
                R = Utils.speye(self.cellDiffStencil.shape[0])

            else:
                r = self.R(self.f_m)
                R = Utils.sdiag(r)

            if self.cell_weights is not None:
                W = (
                    Utils.sdiag(
                        ((Ave * (self.scale * self.cell_weights)))**0.5
                    ) *
                    R
                )

            else:
                W = Utils.sdiag(
                    (Ave * (self.scale * self.regmesh.vol))**0.5
                ) * R

            theta = self.cellDiffStencil * (self.mapping * model)
            dmdx = Utils.matutils.coterminal(theta)

            r = W * dmdx

        else:
            r = self.W * (self.mapping * model)

        mD = self.mapping.deriv(model)
        return mD.T * (self.W.T * r)

    @property
    def _multiplier_pair(self):
        return 'alpha_{orientation}'.format(orientation=self.orientation)

    @property
    def f_m(self):

        if self.mrefInSmooth:

            f_m = self._delta_m(self.model)

        else:
            f_m = self.model

        if self.space == 'spherical':
            theta = self.cellDiffStencil * (self.mapping * f_m)
            dmdx = Utils.matutils.coterminal(theta)

        else:

            if self.gradientType == 'total':
                Ave = getattr(
                    self.regmesh,
                    'aveCC2F{}'.format(self.orientation)
                )

                dmdx = np.abs(self.regmesh.aveFx2CC *
                              self.regmesh.cellDiffxStencil *
                              (self.mapping * f_m)
                              )

                if self.regmesh.dim > 1:

                    dmdx += np.abs(self.regmesh.aveFy2CC *
                                   self.regmesh.cellDiffyStencil *
                                   (self.mapping * f_m)
                                   )

                if self.regmesh.dim > 2:

                    dmdx += np.abs(self.regmesh.aveFz2CC *
                                   self.regmesh.cellDiffzStencil *
                                   (self.mapping * f_m)
                                   )

                dmdx = Ave * dmdx

            else:
                dmdx = self.cellDiffStencil * (self.mapping * f_m)

        return dmdx

    @property
    def cellDiffStencil(self):
        return getattr(
            self.regmesh, 'cellDiff{}Stencil'.format(self.orientation)
        )

    @property
    def W(self):

        Ave = getattr(self.regmesh, 'aveCC2F{}'.format(self.orientation))

        if getattr(self, 'model', None) is None:
            R = Utils.speye(self.cellDiffStencil.shape[0])

        else:
            r = self.R(self.f_m)
            R = Utils.sdiag(r)
        if self.scale is None:
            self.scale = np.ones(self.mapping.shape[0])
        if self.cell_weights is not None:
            return (
                Utils.sdiag(
                    (Ave*(self.scale * self.cell_weights))**0.5
                ) *
                R * self.cellDiffStencil
            )
        else:
            return Utils.sdiag(
                (Ave*(self.scale * self.regmesh.vol))**0.5
                ) * R * self.cellDiffStencil


class Sparse(BaseComboRegularization):
    """
    The regularization is:

    .. math::

        R(m) = \\frac{1}{2}\mathbf{(m-m_\\text{ref})^\\top W^\\top R^\\top R
        W(m-m_\\text{ref})}

    where the IRLS weight

    .. math::

        R = \eta TO FINISH LATER!!!

    So the derivative is straight forward:

    .. math::

        R(m) = \mathbf{W^\\top R^\\top R W (m-m_\\text{ref})}

    The IRLS weights are recomputed after each beta solves.
    It is strongly recommended to do a few Gauss-Newton iterations
    before updating.
    """
    def __init__(
        self, mesh,
        alpha_s=1.0, alpha_x=1.0, alpha_y=1.0, alpha_z=1.0,
        **kwargs
    ):

        objfcts = [
            SparseSmall(mesh=mesh, **kwargs),
            SparseDeriv(mesh=mesh, orientation='x', **kwargs)
        ]

        if mesh.dim > 1:
            objfcts.append(SparseDeriv(mesh=mesh, orientation='y', **kwargs))

        if mesh.dim > 2:
            objfcts.append(SparseDeriv(mesh=mesh, orientation='z', **kwargs))

        super(Sparse, self).__init__(
            mesh=mesh, objfcts=objfcts,
            alpha_s=alpha_s, alpha_x=alpha_x, alpha_y=alpha_y, alpha_z=alpha_z,
            **kwargs
        )

        # Utils.setKwargs(self, **kwargs)

    # Properties
    norms = properties.Array(
        "Norms used to create the sparse regularization",
        default=np.c_[2., 2., 2., 2.], shape={('*', '*')}
    )

    eps_p = properties.Float(
        "Threshold value for the model norm", required=True
        )

    eps_q = properties.Float(
        "Threshold value for the model gradient norm", required=True
        )

    model = properties.Array("current model", dtype=float)

    space = properties.String(
        "type of model", default='linear'
    )

    gradientType = properties.String(
        "type of gradient", default='components'
    )

    scales = properties.Array(
        "General nob for scaling",
        default=np.c_[1., 1., 1., 1.], shape={('*', '*')}
    )
    # Give the option to scale or not
    scaledIRLS = properties.Bool(
        "Scale the gradients of the IRLS norms",
        default=True
    )
    # Save the l2 result during the IRLS
    l2model = None

    @properties.validator('norms')
    def _validate_norms(self, change):
        if change['value'].shape[0] == 1:
            change['value'] = np.kron(
                np.ones((self.regmesh.Pac.shape[1], 1)), change['value']
            )
        elif change['value'].shape[0] > 1:
            assert change['value'].shape[0] == self.regmesh.Pac.shape[1], (
                "Vector of norms must be the size"
                " of active model parameters ({})"
                "The provided vector has length "
                "{}".format(
                    self.regmesh.Pac.shape[0], len(change['value'])
                )
            )

    # Observers
    @properties.observer('norms')
    def _mirror_norms_to_objfcts(self, change):

        self.objfcts[0].norm = change['value'][:, 0]
        for i, objfct in enumerate(self.objfcts[1:]):
            Ave = getattr(objfct.regmesh, 'aveCC2F{}'.format(objfct.orientation))
            objfct.norm = Ave*change['value'][:, i+1]

    @properties.observer('model')
    def _mirror_model_to_objfcts(self, change):
        for objfct in self.objfcts:
            objfct.model = change['value']

    @properties.observer('eps_p')
    def _mirror_eps_p_to_smallness(self, change):
        for objfct in self.objfcts:
            if isinstance(objfct, SparseSmall):
                objfct.epsilon = change['value']

    @properties.observer('eps_q')
    def _mirror_eps_q_to_derivs(self, change):
        for objfct in self.objfcts:
            if isinstance(objfct, SparseDeriv):
                objfct.epsilon = change['value']

    @properties.observer('space')
    def _mirror_space_to_objfcts(self, change):
        for objfct in self.objfcts:
            objfct.space = change['value']

    @properties.observer('gradientType')
    def _mirror_gradientType_to_objfcts(self, change):
        for objfct in self.objfcts:
            objfct.gradientType = change['value']

    @properties.observer('scaledIRLS')
    def _mirror_scaledIRLS_to_objfcts(self, change):
        for objfct in self.objfcts:
            objfct.scaledIRLS = change['value']

    @properties.validator('scales')
    def _validate_scales(self, change):
        if change['value'].shape[0] == 1:
            change['value'] = np.kron(
                np.ones((self.regmesh.Pac.shape[1], 1)), change['value']
            )
        elif change['value'].shape[0] > 1:
            assert change['value'].shape[0] == self.regmesh.Pac.shape[1], (
                "Vector of scales must be the size"
                " of active model parameters ({})"
                "The provided vector has length "
                "{}".format(
                    self.regmesh.Pac.shape[0], len(change['value'])
                )
            )

    # Observers
    @properties.observer('scales')
    def _mirror_scale_to_objfcts(self, change):
        for i, objfct in enumerate(self.objfcts):
            objfct.scale = change['value'][:, i]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
import sys

from SimPEG import Utils
from SimPEG import Props
from SimPEG import Maps

from SimPEG.EM.Base import BaseEMProblem
from SimPEG.EM.Static.DC.FieldsDC import FieldsDC, Fields_CC, Fields_N
from SimPEG.EM.Static.IP import Problem3D_CC as BaseProblem3D_CC
from SimPEG.EM.Static.IP import Problem3D_N as BaseProblem3D_N
from .SurveySIP import Survey, Data
import gc


class BaseSIPProblem(BaseEMProblem):

    sigma = Props.PhysicalProperty(
        "Electrical conductivity (S/m)"
    )

    rho = Props.PhysicalProperty(
        "Electrical resistivity (Ohm m)"
    )

    Props.Reciprocal(sigma, rho)

    eta, etaMap, etaDeriv = Props.Invertible(
        "Electrical Chargeability (V/V)"
    )

    tau, tauMap, tauDeriv = Props.Invertible(
        "Time constant (s)",
        default=0.1
    )

    taui, tauiMap, tauiDeriv = Props.Invertible(
        "Inverse of time constant (1/s)"
    )

    c, cMap, cDeriv = Props.Invertible(
        "Frequency dependency",
        default=0.5
    )

    Props.Reciprocal(tau, taui)

    surveyPair = Survey
    fieldsPair = FieldsDC
    dataPair = Data
    Ainv = None
    _f = None
    actinds = None
    storeJ = False
    _Jmatrix = None
    actMap = None
    n_pulse = 1

    _eta_store = None
    _taui_store = None
    _c_store = None

    @property
    def etaDeriv_store(self):
        if getattr(self, '_etaDeriv_store', None) is None:
            self._etaDeriv_store = self.etaDeriv
        return self._etaDeriv_store

    @property
    def tauiDeriv_store(self):
        if getattr(self, '_tauiDeriv_store', None) is None:
            self._tauiDeriv_store = self.tauiDeriv
        return self._tauiDeriv_store

    @property
    def tauDeriv_store(self):
        if getattr(self, '_tauDeriv_store', None) is None:
            self._tauDeriv_store = self.tauDeriv
        return self._tauDeriv_store

    @property
    def cDeriv_store(self):
        if getattr(self, '_cDeriv_store', None) is None:
            self._cDeriv_store = self.cDeriv
        return self._cDeriv_store

    def get_t_over_tau(self, t):
        taui = self._taui_store
        return (t*taui)

    def get_exponent(self, t):
        c = self._c_store
        t_over_tau = self.get_t_over_tau(t)
        return (t_over_tau)**c

    def get_peta_step_off(self, exponent):
        eta = self._eta_store
        peta = eta*np.exp(-exponent)
        return peta

    def get_peta_pulse_off(self, t):
        """
            Compute pseudo-chargeability from a single pulse waveform
        """
        T = self.survey.T
        exponent_0 = self.get_exponent(t)
        exponent_1 = self.get_exponent(t+T/4.)
        peta = self.get_peta_step_off(exponent_0) - self.get_peta_step_off(exponent_1)
        return peta

    def get_multi_pulse_response(self, t, pulse_func):
        n_pulse = self.survey.n_pulse
        T = self.survey.T
        peta = np.zeros(self._eta_store.shape[0], float, order='C')
        for i_pulse in range (n_pulse):
            factor = (-1)**i_pulse * (n_pulse-i_pulse)
            peta += pulse_func(t+T/2*i_pulse) * factor
        return peta/n_pulse

    def get_peta(self, t):
        n_pulse = self.survey.n_pulse
        if n_pulse == 0:
            exponent = self.get_exponent(t)
            return self.get_peta_step_off(exponent)
        else:
            return self.get_multi_pulse_response(t, self.get_peta_pulse_off)

    def get_peta_eta_deriv_step_off(self, exponent):
        return np.exp(-exponent)

    def get_peta_eta_deriv_pulse_off(self, t):
        """
            Compute derivative of pseudo-chargeability w.r.t eta from a single pulse waveform
        """
        T = self.survey.T
        exponent_0 = self.get_exponent(t)
        exponent_1 = self.get_exponent(t+T/4.)
        peta_eta_deriv = (
            self.get_peta_eta_deriv_step_off(exponent_0) -
            self.get_peta_eta_deriv_step_off(exponent_1)
        )
        return peta_eta_deriv

    def get_peta_eta_deriv(self, t):
        n_pulse = self.survey.n_pulse
        if n_pulse == 0:
            exponent = self.get_exponent(t)
            return self.get_peta_eta_deriv_step_off(exponent)
        else:
            return self.get_multi_pulse_response(
                t, self.get_peta_eta_deriv_pulse_off
            )

    def PetaEtaDeriv(self, t, v, adjoint=False):

        etaDeriv = self.etaDeriv_store

        v = np.array(v, dtype=float)
        dpetadeta = self.get_peta_eta_deriv(t)

        if adjoint:
            if v.ndim == 1:
                return etaDeriv.T * (dpetadeta * v)
            else:
                return etaDeriv.T * (Utils.sdiag(dpetadeta) * v)
        else:
            return dpetadeta * (etaDeriv*v)

    def get_peta_taui_deriv_step_off(self, exponent):
        eta = self._eta_store
        taui = self._taui_store
        c = self._c_store
        peta_taui_deriv = (
            - c * eta / taui * exponent * np.exp(-exponent)
        )
        return peta_taui_deriv

    def get_peta_taui_deriv_pulse_off(self, t):
        """
            Compute derivative of pseudo-chargeability w.r.t eta from a single pulse waveform
        """
        T = self.survey.T
        exponent_0 = self.get_exponent(t)
        exponent_1 = self.get_exponent(t+T/4.)
        peta_taui_deriv = (
            self.get_peta_taui_deriv_step_off(exponent_0) -
            self.get_peta_taui_deriv_step_off(exponent_1)
        )
        return peta_taui_deriv

    def get_peta_taui_deriv(self, t):
        n_pulse = self.survey.n_pulse
        if n_pulse == 0:
            exponent = self.get_exponent(t)
            return self.get_peta_taui_deriv_step_off(exponent)
        else:
            return self.get_multi_pulse_response(
                t, self.get_peta_taui_deriv_pulse_off
            )

    def PetaTauiDeriv(self, t, v, adjoint=False):
        v = np.array(v, dtype=float)
        tauiDeriv = self.tauiDeriv_store
        dpetadtaui = self.get_peta_taui_deriv(t)

        if adjoint:
            if v.ndim == 1:
                return tauiDeriv.T * (dpetadtaui*v)
            else:
                return tauiDeriv.T * (Utils.sdiag(dpetadtaui)*v)
        else:
            return dpetadtaui * (tauiDeriv*v)

    def get_peta_c_deriv_step_off(self, exponent, t_over_tau):
        eta = self._eta_store
        peta_c_deriv = (
            - eta * (exponent)*np.exp(-exponent) * np.log(t_over_tau)
        )
        return peta_c_deriv

    def get_peta_c_deriv_pulse_off(self, t):
        """
            Compute derivative of pseudo-chargeability w.r.t eta from a single pulse waveform
        """
        T = self.survey.T
        exponent_0 = self.get_exponent(t)
        exponent_1 = self.get_exponent(t+T/4.)
        t_over_tau_0 = self.get_t_over_tau(t)
        t_over_tau_1 = self.get_t_over_tau(t+T/4.)

        peta_c_deriv = (
            self.get_peta_c_deriv_step_off(exponent_0, t_over_tau_0) -
            self.get_peta_c_deriv_step_off(exponent_1, t_over_tau_1)
        )
        return peta_c_deriv

    def get_peta_c_deriv(self, t):
        n_pulse = self.survey.n_pulse
        if n_pulse == 0:
            exponent = self.get_exponent(t)
            t_over_tau = self.get_t_over_tau(t)
            return self.get_peta_c_deriv_step_off(exponent, t_over_tau)
        else:
            return self.get_multi_pulse_response(
                t, self.get_peta_c_deriv_pulse_off
            )

    def PetaCDeriv(self, t, v, adjoint=False):
        v = np.array(v, dtype=float)
        cDeriv = self.cDeriv_store
        dpetadc = self.get_peta_c_deriv(t)
        if adjoint:
            if v.ndim == 1:
                return cDeriv.T * (dpetadc*v)
            else:
                return cDeriv.T * (Utils.sdiag(dpetadc)*v)
        else:
            return dpetadc * (cDeriv*v)

    def fields(self, m):

        if self.verbose:
            print(">> Compute DC fields")
        if self._f is None:

            if self.verbose:
                print (">> Compute DC fields")

            self._f = self.fieldsPair(self.mesh, self.survey)
            if self.Ainv is None:
                A = self.getA()
                self.Ainv = self.Solver(A, **self.solverOpts)
            RHS = self.getRHS()
            u = self.Ainv * RHS
            Srcs = self.survey.srcList
            self._f[Srcs, self._solutionType] = u

            # Compute DC voltage
            if self.data_type == 'apparent_chargeability':
                if self.verbose is True:
                    print(">> Data type is apparaent chargeability")
                for src in self.survey.srcList:
                    for rx in src.rxList:
                        rx._dc_voltage = rx.eval(src, self.mesh, self._f)
                        rx.data_type = self.data_type
                        rx._Ps = {}

        self.survey._pred = self.forward(m, f=self._f)

        return self._f

    # @profile
    def getJ(self, m, f=None):
        """
            Generate Full sensitivity matrix
        """

        if self._Jmatrix is not None:
            return self._Jmatrix
        else:
            if self.verbose:
                print("Calculating J and storing")
            if f is None:
                f = self.fields(m)

            Jt = np.zeros(
                (self.actMap.nP, int(self.survey.nD/self.survey.times.size)),
                order='F'
            )
            istrt = int(0)
            iend = int(0)

            for isrc, src in enumerate(self.survey.srcList):
                if self.verbose:
                    sys.stdout.write(
                        ("\r %d / %d") % (isrc+1, self.survey.nSrc)
                    )
                    sys.stdout.flush()
                u_src = f[src, self._solutionType]
                for rx in src.rxList:
                    P = rx.getP(self.mesh, rx.projGLoc(f)).toarray()
                    ATinvdf_duT = self.Ainv * (P.T)
                    dA_dmT = self.getADeriv(u_src, ATinvdf_duT, adjoint=True)
                    iend = istrt + rx.nD
                    if rx.nD == 1:
                        Jt[:, istrt] = -dA_dmT
                    else:
                        Jt[:, istrt:iend] = -dA_dmT
                    istrt += rx.nD

            self._Jmatrix = Jt.T
            collected = gc.collect()
            if self.verbose:
                collected = gc.collect()
                print(
                    "Garbage collector: collected %d objects." % (collected)
                )
            # clean field object
            self._f = []
            # clean all factorization
            if self.Ainv is not None:
                self.Ainv.clean()

            return self._Jmatrix

    def getJtJdiag(self, m, Wd):
        """
        Compute JtJ using adjoint problem. Still we never form
        JtJ
        """
        if self.verbose:
            print (">> Compute trace(JtJ)")
        ntime = len(self.survey.times)
        JtJdiag = np.zeros_like(m)
        J = self.getJ(m, f=None)
        wd = (Wd.diagonal()).reshape(
            (self.survey.n_locations, ntime), order='F'
        )
        for tind in range(ntime):
            t = self.survey.times[tind]
            Jtv = self.actMap.P*J.T*Utils.sdiag(wd[:, tind])
            JtJdiag += (
                (self.PetaEtaDeriv(t, Jtv, adjoint=True)**2).sum(axis=1) +
                (self.PetaTauiDeriv(t, Jtv, adjoint=True)**2).sum(axis=1) +
                (self.PetaCDeriv(t, Jtv, adjoint=True)**2).sum(axis=1)
            )
        return JtJdiag

    # @profile
    def forward(self, m, f=None):

        if self.verbose:
            print ('>> Compute predicted data')

        self.model = m

        self._eta_store = self.eta
        self._taui_store = self.taui
        self._c_store = self.c

        Jv = []

        # When sensitivity matrix is stored
        if self.storeJ:
            J = self.getJ(m, f=f)

            ntime = len(self.survey.times)

            self.model = m
            for tind in range(ntime):
                Jv.append(
                    J.dot(
                        self.actMap.P.T*self.get_peta(self.survey.times[tind]))
                    )
            return self.sign * np.hstack(Jv)

        # Do not store sensitivity matrix (memory-wise efficient)
        else:

            if f is None:
                f = self.fields(m)

            # A = self.getA()
            for tind in range(len(self.survey.times)):
                # Pseudo-chareability
                t = self.survey.times[tind]
                v = self.get_peta(t)
                for src in self.survey.srcList:
                    u_src = f[src, self._solutionType]  # solution vector
                    dA_dm_v = self.getADeriv(u_src, v)
                    dRHS_dm_v = self.getRHSDeriv(src, v)
                    du_dm_v = self.Ainv * (- dA_dm_v + dRHS_dm_v)
                    for rx in src.rxList:
                        timeindex = rx.getTimeP(self.survey.times)
                        if timeindex[tind]:
                            df_dmFun = getattr(
                                f, '_{0!s}Deriv'.format(rx.projField),
                                None
                                )
                            df_dm_v = df_dmFun(src, du_dm_v, v, adjoint=False)
                            Jv.append(
                                rx.evalDeriv(src, self.mesh, f, df_dm_v)
                                )

            return self.sign*np.hstack(Jv)

    # @profile
    def Jvec(self, m, v, f=None):

        self.model = m

        Jv = []

        # When sensitivity matrix is stored
        if self.storeJ:
            J = self.getJ(m, f=f)
            ntime = len(self.survey.times)

            for tind in range(ntime):

                t = self.survey.times[tind]
                v0 = self.PetaEtaDeriv(t, v)
                v1 = self.PetaTauiDeriv(t, v)
                v2 = self.PetaCDeriv(t, v)
                PTv = self.actMap.P.T*(v0+v1+v2)
                Jv.append(J.dot(PTv))

            return self.sign * np.hstack(Jv)

        # Do not store sensitivity matrix (memory-wise efficient)
        else:

            if f is None:
                f = self.fields(m)

            for tind in range(len(self.survey.times)):

                t = self.survey.times[tind]
                v0 = self.PetaEtaDeriv(t, v)
                v1 = self.PetaTauiDeriv(t, v)
                v2 = self.PetaCDeriv(t, v)

                for src in self.survey.srcList:
                    u_src = f[src, self._solutionType]  # solution vector
                    dA_dm_v = self.getADeriv(u_src, v0+v1+v2)
                    dRHS_dm_v = self.getRHSDeriv(src, v0+v1+v2)
                    du_dm_v = self.Ainv * (
                        - dA_dm_v + dRHS_dm_v
                        )

                    for rx in src.rxList:
                        # Assume same # of time
                        # timeindex = rx.getTimeP(self.survey.times)
                        # if timeindex[tind]:
                        Jv_temp = (
                            rx.evalDeriv(src, self.mesh, f, du_dm_v)
                            )
                        Jv.append(Jv_temp)

            return self.sign*np.hstack(Jv)

    def Jtvec(self, m, v, f=None):

        self.model = m

        # When sensitivity matrix is stored
        if self.storeJ:
            J = self.getJ(m, f=f)
            ntime = len(self.survey.times)
            Jtvec = np.zeros(m.size)
            v = v.reshape((int(self.survey.nD/ntime), ntime), order="F")

            for tind in range(ntime):
                t = self.survey.times[tind]
                Jtv = self.actMap.P*J.T.dot(v[:, tind])
                Jtvec += (
                    self.PetaEtaDeriv(t, Jtv, adjoint=True) +
                    self.PetaTauiDeriv(t, Jtv, adjoint=True) +
                    self.PetaCDeriv(t, Jtv, adjoint=True)
                    )

            return self.sign * Jtvec

        # Do not store sensitivity matrix (memory-wise efficient)
        else:

            if f is None:
                f = self.fields(m)

            # Ensure v is a data object.
            if not isinstance(v, self.dataPair):
                v = self.dataPair(self.survey, v)

            Jtv = np.zeros(m.size, dtype=float)
            n_time = len(self.survey.times)
            du_dmT = np.zeros((self.mesh.nC, n_time), dtype=float, order='F')

            for tind in range(n_time):
                t = self.survey.times[tind]

                for src in self.survey.srcList:
                    u_src = f[src, self._solutionType]
                    for rx in src.rxList:

                        # Ignore case when each rx has different # of times

                        # timeindex = rx.getTimeP(self.survey.times)
                        # if timeindex[tind]:
                        # wrt f, need possibility wrt m
                        PTv = rx.evalDeriv(
                            src, self.mesh, f, v[src, rx, t], adjoint=True
                        )
                        df_duTFun = getattr(
                            f, '_{0!s}Deriv'.format(rx.projField), None
                        )
                        df_duT, _ = df_duTFun(
                            src, None, PTv, adjoint=True
                        )
                        ATinvdf_duT = self.Ainv * df_duT
                        dA_dmT = self.getADeriv(
                            u_src, ATinvdf_duT, adjoint=True
                        )
                        # Unecessary at the moment

                        # dRHS_dmT = self.getRHSDeriv(
                        #     src, ATinvdf_duT, adjoint=True
                        # )
                        # du_dmT[:, tind] = -dA_dmT + dRHS_dmT

                        du_dmT[:, tind] += -self.getADeriv(
                            u_src, ATinvdf_duT, adjoint=True
                        )

                Jtv += (
                    self.PetaEtaDeriv(
                        self.survey.times[tind], du_dmT[:, tind],
                        adjoint=True
                    ) +
                    self.PetaTauiDeriv(
                        self.survey.times[tind], du_dmT[:, tind],
                        adjoint=True
                    ) +
                    self.PetaCDeriv(
                        self.survey.times[tind], du_dmT[:, tind],
                        adjoint=True
                    )
                )

            return self.sign*Jtv

    def getSourceTerm(self):
        """
        takes concept of source and turns it into a matrix
        """
        """
        Evaluates the sources, and puts them in matrix form

        :rtype: (numpy.ndarray, numpy.ndarray)
        :return: q (nC or nN, nSrc)
        """

        Srcs = self.survey.srcList

        if self._formulation == 'EB':
            n = self.mesh.nN
            # return NotImplementedError

        elif self._formulation == 'HJ':
            n = self.mesh.nC

        q = np.zeros((n, len(Srcs)))

        for i, src in enumerate(Srcs):
            q[:, i] = src.eval(self)
        return q

    @property
    def deleteTheseOnModelUpdate(self):
        toDelete = [
            '_etaDeriv_store', '_tauiDeriv_store', '_cDeriv_store',
            '_tauDeriv_store'
        ]
        return toDelete

    @property
    def MfRhoDerivMat(self):
        """
        Derivative of MfRho with respect to the model
        """
        if getattr(self, '_MfRhoDerivMat', None) is None:
            if self.storeJ:
                drho_dlogrho = Utils.sdiag(self.rho)*self.actMap.P
            else:
                drho_dlogrho = Utils.sdiag(self.rho)
            self._MfRhoDerivMat = self.mesh.getFaceInnerProductDeriv(
                np.ones(self.mesh.nC)
            )(np.ones(self.mesh.nF)) * drho_dlogrho
        return self._MfRhoDerivMat

    def MfRhoIDeriv(self, u, v, adjoint=False):
        """
            Derivative of :code:`MfRhoI` with respect to the model.
        """
        dMfRhoI_dI = -self.MfRhoI**2

        if self.storeInnerProduct:
            if adjoint:
                return (
                    self.MfRhoDerivMat.T * (
                        Utils.sdiag(u) * (dMfRhoI_dI.T * v)
                    )
                )
            else:
                return dMfRhoI_dI * (Utils.sdiag(u) * (self.MfRhoDerivMat*v))
        else:
            if self.storeJ:
                drho_dlogrho = Utils.sdiag(self.rho)*self.actMap.P
            else:
                drho_dlogrho = Utils.sdiag(self.rho)
            dMf_drho = self.mesh.getFaceInnerProductDeriv(self.rho)(u)
            if adjoint:
                return drho_dlogrho.T * (dMf_drho.T * (dMfRhoI_dI.T*v))
            else:
                return dMfRhoI_dI * (dMf_drho * (drho_dlogrho*v))

    @property
    def MeSigmaDerivMat(self):
        """
        Derivative of MeSigma with respect to the model
        """
        if getattr(self, '_MeSigmaDerivMat', None) is None:
            if self.storeJ:
                dsigma_dlogsigma = Utils.sdiag(self.sigma)*self.actMap.P
            else:
                dsigma_dlogsigma = Utils.sdiag(self.sigma)
            self._MeSigmaDerivMat = self.mesh.getEdgeInnerProductDeriv(
                np.ones(self.mesh.nC)
            )(np.ones(self.mesh.nE)) * dsigma_dlogsigma
        return self._MeSigmaDerivMat

    # TODO: This should take a vector
    def MeSigmaDeriv(self, u, v, adjoint=False):
        """
        Derivative of MeSigma with respect to the model times a vector (u)
        """
        if self.storeInnerProduct:
            if adjoint:
                return self.MeSigmaDerivMat.T * (Utils.sdiag(u)*v)
            else:
                return Utils.sdiag(u)*(self.MeSigmaDerivMat * v)
        else:
            if self.storeJ:
                dsigma_dlogsigma = Utils.sdiag(self.sigma)*self.actMap.P
            else:
                dsigma_dlogsigma = Utils.sdiag(self.sigma)
            if adjoint:
                return (
                    dsigma_dlogsigma.T * (
                        self.mesh.getEdgeInnerProductDeriv(self.sigma)(u).T * v
                    )
                )
            else:
                return (
                    self.mesh.getEdgeInnerProductDeriv(self.sigma)(u) *
                    (dsigma_dlogsigma * v)
                )


class Problem3D_CC(BaseSIPProblem, BaseProblem3D_CC):

    _solutionType = 'phiSolution'
    _formulation = 'HJ'  # CC potentials means J is on faces
    fieldsPair = Fields_CC
    sign = 1.
    bc_type = 'Neumann'

    def __init__(self, mesh, **kwargs):
        BaseSIPProblem.__init__(self, mesh, **kwargs)
        self.setBC()
        self.n = self.mesh.nC
        if self.storeJ:
            if self.actinds is None:
                print ("You did not put Active indices")
                print ("So, set actMap = IdentityMap(mesh)")
                self.actinds = np.ones(mesh.nC, dtype=bool)

            self.actMap = Maps.InjectActiveCells(mesh, self.actinds, 0.)


class Problem3D_N(BaseSIPProblem, BaseProblem3D_N):

    _solutionType = 'phiSolution'
    _formulation = 'EB'  # N potentials means B is on faces
    fieldsPair = Fields_N
    sign = -1.

    def __init__(self, mesh, **kwargs):
        BaseSIPProblem.__init__(self, mesh, **kwargs)
        self.n = self.mesh.nN
        if self.storeJ:
            if self.actinds is None:
                print ("You did not put Active indices")
                print ("So, set actMap = IdentityMap(mesh)")
                self.actinds = np.ones(mesh.nC, dtype=bool)

            self.actMap = Maps.InjectActiveCells(mesh, self.actinds, 0.)



from .ProblemIP import Problem3D_CC, Problem3D_N
from .ProblemIP_2D import Problem2D_CC, Problem2D_N, BaseIPProblem_2D
from .SurveyIP import Survey, from_dc_to_ip_survey
from .Run import run_inversion



""" module SimPEG.EM.NSEM.Utils

Collection of utilities that are usefull for the NSEM problem

NOTE: These utilities are not well test, use with care

"""
from __future__ import absolute_import

from .MT1Dsolutions import get1DEfields  # Add the names of the functions
from .MT1Danalytic import getEHfields, getImpedance
from .dataUtils import (appResPhs, rec_to_ndarr, rotate_data,
                        skindepth, makeAnalyticSolution, plotMT1DModelData,
                        plotImpAppRes, printTime, convert3Dto1Dobject,
                        resample_data, extract_data_info)
from .ediFilesUtils import (EDIimporter,
                            _findLatLong, _findLine, _findEDIcomp)
from .testUtils import (getAppResPhs, setup1DSurvey, setupSimpegNSEM_ePrimSec,
                        random, halfSpace, blockInhalfSpace, twoLayer)



from SimPEG import Problem, Utils, Maps, Mesh
from SimPEG.EM.Base import BaseEMProblem
from SimPEG.EM.Static.DC.FieldsDC import FieldsDC, Fields_CC
from SimPEG.EM.Static.DC import Survey, BaseDCProblem, Problem3D_CC
from SimPEG.Utils import sdiag
import numpy as np
import scipy.sparse as sp
from SimPEG.Utils import Zero
from SimPEG.EM.Static.DC import getxBCyBC_CC
from SimPEG import Props
import properties


class BaseSPProblem(BaseDCProblem):

    h, hMap, hDeriv = Props.Invertible(
        "Hydraulic Head (m)"
    )

    q, qMap, qDeriv = Props.Invertible(
        "Streaming current source (A/m^3)"
    )

    jsx, jsxMap, jsxDeriv = Props.Invertible(
        "Streaming current density in x-direction (A/m^2)"
    )

    jsy, jsyMap, jsyDeriv = Props.Invertible(
        "Streaming current density in y-direction (A/m^2)"
    )

    jsz, jszMap, jszDeriv = Props.Invertible(
        "Streaming current density in z-direction (A/m^2)"
    )

    sigma = Props.PhysicalProperty(
        "Electrical conductivity (S/m)"
    )

    rho = Props.PhysicalProperty(
        "Electrical resistivity (Ohm m)"
    )

    Props.Reciprocal(sigma, rho)

    modelType = None
    surveyPair = Survey
    fieldsPair = FieldsDC

    @property
    def deleteTheseOnModelUpdate(self):
        toDelete = []
        return toDelete

    def evalq(self, Qv, vel):
        MfQviI = self.mesh.getFaceInnerProduct(1./Qv, invMat=True)
        Mf = self.mesh.getFaceInnerProduct()
        return self.Div*(Mf*(MfQviI*vel))


class Problem_CC(BaseSPProblem, Problem3D_CC):

    _solutionType = 'phiSolution'
    _formulation = 'HJ'  # CC potentials means J is on faces
    fieldsPair = Fields_CC
    modelType = None
    bc_type = "Mixed"
    coordinate_system = properties.StringChoice(
        "Type of coordinate system we are regularizing in",
        choices=['cartesian', 'spherical'],
        default='cartesian'
    )

    def __init__(self, mesh, **kwargs):
        BaseSPProblem.__init__(self, mesh, **kwargs)
        self.setBC()

    def getADeriv(self, u, v, adjoint= False):
        # We assume conductivity is known
        return Zero()

    def getRHSDeriv(self, src, v, adjoint=False):
        """
        Derivative of the right hand side with respect to the model
        """
        return src.evalDeriv(self, v=v, adjoint=adjoint)


class Problem_CC_Jstore(Problem_CC):
    """docstring for Problem_CC_jstore"""

    _S = None

    @property
    def G(self):
        """
            Inverse of :code:`_G`
        """
        if getattr(self, '_G', None) is None:
            A = self.getA()
            self.Ainv = self.Solver(A, **self.solverOpts)
            src = self.survey.srcList[0]
            rx = src.rxList[0]
            P = rx.getP(self.mesh, "CC").toarray()
            src = self.survey.srcList[0]
            self._G = (self.Ainv * P.T).T * src.evalDeriv(
                self, v=Utils.sdiag(np.ones_like(self.model))
            )
            self.Ainv.clean()
            del self.Ainv
        return self._G

    def getJ(self, m, f=None):

        if self.coordinate_system == 'cartesian':
            return self.G
        else:
            self.model = m
            return self.G * self.S

    def Jvec(self, m, v, f=None):

        self.model = m

        if self.coordinate_system == 'cartesian':
            return self.G.dot(v)
        else:
            return np.dot(self.G, self.S.dot(v))

    def Jtvec(self, m, v, f=None):

        self.model = m

        if self.coordinate_system == 'cartesian':
            return self.G.T.dot(v)
        else:
            return self.S.T*(self.G.T.dot(v))

    @Utils.count
    def fields(self, m):

        self.model = m

        if self.coordinate_system == 'spherical':
            m = Utils.matutils.atp2xyz(m)

        return self.G.dot(m)

    @property
    def S(self):
        """
            Derivatives for the spherical transformation
        """
        if getattr(self, '_S', None) is None:
            if self.verbose:
                print ("Compute S")
            if self.model is None:
                raise Exception('Requires a model')

            # Assume it is vector model in spherical coordinates
            nC = int(self.model.shape[0]/3)

            a = self.model[:nC]
            t = self.model[nC:2*nC]
            p = self.model[2*nC:]

            Sx = sp.hstack([sp.diags(np.cos(t)*np.cos(p), 0),
                            sp.diags(-a*np.sin(t)*np.cos(p), 0),
                            sp.diags(-a*np.cos(t)*np.sin(p), 0)])

            Sy = sp.hstack([sp.diags(np.cos(t)*np.sin(p), 0),
                            sp.diags(-a*np.sin(t)*np.sin(p), 0),
                            sp.diags(a*np.cos(t)*np.cos(p), 0)])

            Sz = sp.hstack([sp.diags(np.sin(t), 0),
                            sp.diags(a*np.cos(t), 0),
                            sp.csr_matrix((nC, nC))])

            self._S = sp.vstack([Sx, Sy, Sz])

        return self._S

    @property
    def deleteTheseOnModelUpdate(self):
        toDelete = super(BaseDCProblem, self).deleteTheseOnModelUpdate
        if self._S is not None:
            toDelete += ['_S']
        return toDelete


class SurveySP_store(Survey):
    @Utils.count
    @Utils.requires('prob')
    def dpred(self, m=None, f=None):
        return self.prob.fields(m)





from __future__ import division
import numpy as np

from discretize.utils import (
    Zero, Identity, mkvc, sdiag, sdInv, speye, kron3, spzeros, ddx, av,
    av_extrap, ndgrid, ind2sub, sub2ind, getSubArray, inv3X3BlockDiagonal,
    inv2X2BlockDiagonal, TensorType, makePropertyTensor, invPropertyTensor,
)


def avExtrap(**kwargs):
    raise Exception("avExtrap has been depreciated. Use av_extrap instead.")


def diagEst(matFun, n, k=None, approach='Probing'):
    """
        Estimate the diagonal of a matrix, A. Note that the matrix may be a
        function which returns A times a vector.

        Three different approaches have been implemented:

        1. Probing: cyclic permutations of vectors with 1's and 0's (default)
        2. Ones: random +/- 1 entries
        3. Random: random vectors

        :param callable matFun: takes a (numpy.ndarray) and multiplies it by a matrix to estimate the diagonal
        :param int n: size of the vector that should be used to compute matFun(v)
        :param int k: number of vectors to be used to estimate the diagonal
        :param str approach: approach to be used for getting vectors
        :rtype: numpy.ndarray
        :return: est_diag(A)

        Based on Saad http://www-users.cs.umn.edu/~saad/PDF/umsi-2005-082.pdf,
        and http://www.cita.utoronto.ca/~niels/diagonal.pdf
    """

    if type(matFun).__name__ == 'ndarray':
        A = matFun

        def matFun(v):
            return A.dot(v)

    if k is None:
        k = np.floor(n/10.)

    if approach.upper() == 'ONES':
        def getv(n, i=None):
            v = np.random.randn(n)
            v[v < 0] = -1.
            v[v >= 0] = 1.
            return v

    elif approach.upper() == 'RANDOM':
        def getv(n, i=None):
            return np.random.randn(n)

    else:  # if approach == 'Probing':
        def getv(n, i):
            v = np.zeros(n)
            v[i:n:k] = 1.
            return v

    Mv = np.zeros(n)
    vv = np.zeros(n)

    for i in range(0, k):
        vk = getv(n, i)
        Mv += matFun(vk)*vk
        vv += vk*vk

    d = Mv/vv

    return d


def uniqueRows(M):
    b = np.ascontiguousarray(M).view(np.dtype(
        (np.void, M.dtype.itemsize * M.shape[1]))
    )
    _, unqInd = np.unique(b, return_index=True)
    _, invInd = np.unique(b, return_inverse=True)
    unqM = M[unqInd]
    return unqM, unqInd, invInd


def cartesian2spherical(m):
    """ Convert from cartesian to spherical """

    # nC = int(len(m)/3)

    x = m[:, 0]
    y = m[:, 1]
    z = m[:, 2]

    a = (x**2. + y**2. + z**2.)**0.5

    t = np.zeros_like(x)
    t[a > 0] = np.arcsin(z[a > 0]/a[a > 0])

    p = np.zeros_like(x)
    p[a > 0] = np.arctan2(y[a > 0], x[a > 0])

    m_atp = np.r_[a, t, p]

    return m_atp


def spherical2cartesian(m):
    """ Convert from spherical to cartesian """

    a = m[:, 0] + 1e-8
    t = m[:, 1]
    p = m[:, 2]

    m_xyz = np.r_[a*np.cos(t)*np.cos(p),
                  a*np.cos(t)*np.sin(p),
                  a*np.sin(t)]

    return m_xyz


def dip_azimuth2cartesian(dip, azm_N):
    """
    dip_azimuth2cartesian(dip,azm_N)

    Function converting degree angles for dip and azimuth from north to a
    3-components in cartesian coordinates.

    INPUT
    dip     : Value or vector of dip from horizontal in DEGREE
    azm_N   : Value or vector of azimuth from north in DEGREE

    OUTPUT
    M       : [n-by-3] Array of xyz components of a unit vector in cartesian

    Created on Dec, 20th 2015

    @author: dominiquef
    """

    azm_N = np.asarray(azm_N)
    dip = np.asarray(dip)

    # Number of elements
    nC = azm_N.size

    M = np.zeros((nC, 3))

    # Modify azimuth from North to cartesian-X
    azm_X = (450. - np.asarray(azm_N)) % 360.
    inc = -np.deg2rad(np.asarray(dip))
    dec = np.deg2rad(azm_X)

    M[:, 0] = np.cos(inc) * np.cos(dec)
    M[:, 1] = np.cos(inc) * np.sin(dec)
    M[:, 2] = np.sin(inc)

    return M


def coterminal(theta):
    """
    Compute coterminal angle so that [-pi < theta < pi]
    """

    sub = theta[np.abs(theta) >= np.pi]
    sub = -np.sign(sub) * (2*np.pi-np.abs(sub))

    theta[np.abs(theta) >= np.pi] = sub

    return theta



import numpy as np
from scipy.interpolate import LinearNDInterpolator, NearestNDInterpolator
import matplotlib.pyplot as plt


def plot2Ddata(
    xyz, data, vec=False, nx=100, ny=100,
    ax=None, mask=None, level=False, figname=None,
    ncontour=10, dataloc=False, contourOpts={},
    levelOpts={}, scale="linear", clim=None,
    method='linear'
):
    """

        Take unstructured xy points, interpolate, then plot in 2D

        :param numpy.ndarray xyz: data locations
        :param numpy.ndarray data: data values
        :param bool vec: plot streamplot?
        :param float nx: number of x grid locations
        :param float ny: number of y grid locations
        :param matplotlib.axes ax: axes
        :param boolean numpy.ndarray mask: mask for the array
        :param boolean level: boolean to plot (or not)
                                :meth:`matplotlib.pyplot.contour`
        :param string figname: figure name
        :param float ncontour: number of :meth:`matplotlib.pyplot.contourf`
                                contours
        :param bool dataloc: plot the data locations
        :param dict controuOpts: :meth:`matplotlib.pyplot.contourf` options
        :param dict levelOpts: :meth:`matplotlib.pyplot.contour` options
        :param numpy.ndarray clim: colorbar limits
        :param str method: interpolation method, either 'linear' or 'nearest'

    """

    # Error checking and set vmin, vmax
    vlimits = [None, None]

    if clim is not None:
        vlimits = [np.min(clim), np.max(clim)]

    for i, key in enumerate(["vmin", "vmax"]):
        if key in contourOpts.keys():
            if vlimits[i] is None:
                vlimits[i] = contourOpts.pop(key)
            else:
                if not np.isclose(contourOpts[key], vlimits[i]):
                    raise Exception(
                        "The values provided in the colorbar limit, clim {} "
                        "does not match the value of {} provided in the "
                        "contourOpts: {}. Only one value should be provided or "
                        "the two values must be equal.".format(
                            vlimits[i], key, contourOpts[key]
                        )
                    )
                contourOpts.pop(key)
    vmin, vmax = vlimits[0], vlimits[1]

    # create a figure if it doesn't exist
    if ax is None:
        fig = plt.figure()
        ax = plt.subplot(111)

    # interpolate data to grid locations
    xmin, xmax = xyz[:, 0].min(), xyz[:, 0].max()
    ymin, ymax = xyz[:, 1].min(), xyz[:, 1].max()
    x = np.linspace(xmin, xmax, nx)
    y = np.linspace(ymin, ymax, ny)
    X, Y = np.meshgrid(x, y)
    xy = np.c_[X.flatten(), Y.flatten()]

    if vec is False:
        if method == 'nearest':
            F = NearestNDInterpolator(xyz[:, :2], data)
        else:
            F = LinearNDInterpolator(xyz[:, :2], data)
        DATA = F(xy)
        DATA = DATA.reshape(X.shape)

        # Levels definitions
        dataselection = np.logical_and(
            ~np.isnan(DATA),
            np.abs(DATA) != np.inf
        )

        if scale == "log":
            DATA = np.log10(abs(DATA))

            vmin = np.log10(vmin) if vmin is not None else vmin
            vmax = np.log10(vmax) if vmax is not None else vmax

        vmin = DATA[dataselection].min() if vmin is None else vmin
        vmax = DATA[dataselection].max() if vmax is None else vmax

        vstep = np.abs((vmin-vmax)/(ncontour+1))
        levels = np.arange(vmin, vmax+vstep, vstep)
        if DATA[dataselection].min() < levels.min():
                levels = np.r_[DATA[dataselection].min(), levels]
        if DATA[dataselection].max() > levels.max():
                levels = np.r_[levels, DATA[dataselection].max()]

        if mask is not None:
            Fmask = NearestNDInterpolator(xyz[:, :2], mask)
            MASK = Fmask(xy)
            MASK = MASK.reshape(X.shape)
            DATA = np.ma.masked_array(DATA, mask=MASK)

        cont = ax.contourf(
            X, Y, DATA, levels=levels, vmin=vmin, vmax=vmax,
            **contourOpts
        )
        if level:
            CS = ax.contour(X, Y, DATA, levels=levels, **levelOpts)

    else:
        # Assume size of data is (N,2)
        datax = data[:, 0]
        datay = data[:, 1]
        if method == 'nearest':
            Fx = NearestNDInterpolator(xyz[:, :2], datax)
            Fy = NearestNDInterpolator(xyz[:, :2], datay)
        else:
            Fx = LinearNDInterpolator(xyz[:, :2], datax)
            Fy = LinearNDInterpolator(xyz[:, :2], datay)
        DATAx = Fx(xy)
        DATAy = Fy(xy)
        DATA = np.sqrt(DATAx**2+DATAy**2).reshape(X.shape)
        DATAx = DATAx.reshape(X.shape)
        DATAy = DATAy.reshape(X.shape)
        if scale == "log":
            DATA = np.log10(abs(DATA))

        # Levels definitions
        dataselection = np.logical_and(
            ~np.isnan(DATA),
            np.abs(DATA) != np.inf
            )


        # set vmin, vmax
        vmin = DATA[dataselection].min() if vmin is None else vmin
        vmax = DATA[dataselection].max() if vmax is None else vmax
        if scale == "log":
            if vmin <= 0 or vmax <= 0:
                raise Exception(
                    "All values must be strictly positive in order to use the log-scale"
                )
            vmin = np.log10(vmin)
            vmax = np.log10(vmax)

        vstep = np.abs((vmin-vmax)/(ncontour+1))
        levels = np.arange(vmin, vmax+vstep, vstep)
        if DATA[dataselection].min() < levels.min():
                levels = np.r_[DATA[dataselection].min(), levels]
        if DATA[dataselection].max() > levels.max():
                levels = np.r_[levels, DATA[dataselection].max()]

        if mask is not None:
            Fmask = NearestNDInterpolator(xyz[:, :2], mask)
            MASK = Fmask(xy)
            MASK = MASK.reshape(X.shape)
            DATA = np.ma.masked_array(DATA, mask=MASK)

        cont = ax.contourf(
            X, Y, DATA, levels=levels,
            vmin=vmin, vmax=vmax,
            **contourOpts
        )
        ax.streamplot(X, Y, DATAx, DATAy, color="w")
        if level:
            CS = ax.contour(X, Y, DATA, levels=levels, **levelOpts)

    if dataloc:
        ax.plot(xyz[:, 0], xyz[:, 1], 'k.', ms=2)
    plt.gca().set_aspect('equal', adjustable='box')
    if figname:
        plt.axis("off")
        fig.savefig(figname, dpi=200)
    if level:
        return cont, ax, CS
    else:
        return cont, ax


def plotLayer(sig, LocSigZ, xscale='log', ax=None,
              showlayers=False, xlim=None, **kwargs):
    """Plot a layered earth model"""
    sigma = np.repeat(sig, 2, axis=0)
    z = np.repeat(LocSigZ[1:], 2, axis=0)
    z = np.r_[LocSigZ[0], z, LocSigZ[-1]]

    if xlim is None:
        sig_min = sig.min()*0.5
        sig_max = sig.max()*2
    else:
        sig_min, sig_max = xlim

    if xscale == 'linear' and sig.min() == 0.:
        if xlim is None:
            sig_min = -sig.max()*0.5
            sig_max = sig.max()*2

    if ax is None:
        plt.xscale(xscale)
        plt.xlim(sig_min, sig_max)
        plt.ylim(z.min(), z.max())
        plt.xlabel('Conductivity (S/m)', fontsize=14)
        plt.ylabel('Depth (m)', fontsize=14)
        plt.ylabel('Depth (m)', fontsize=14)
        if showlayers is True:
            for locz in LocSigZ:
                plt.plot(
                    np.linspace(sig_min, sig_max, 100),
                    np.ones(100)*locz, 'b--', lw=0.5
                )
        return plt.plot(sigma, z, 'k-', **kwargs)

    else:
        ax.set_xscale(xscale)
        ax.set_xlim(sig_min, sig_max)
        ax.set_ylim(z.min(), z.max())
        ax.set_xlabel('Conductivity (S/m)', fontsize=14)
        ax.set_ylabel('Depth (m)', fontsize=14)
        if showlayers is True:
            for locz in LocSigZ:
                ax.plot(
                    np.linspace(sig_min, sig_max, 100),
                    np.ones(100)*locz, 'b--', lw=0.5
                )
        return ax.plot(sigma, z, 'k-', **kwargs)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import numpy as np
import scipy.sparse as sp

import SimPEG as simpeg

from SimPEG import Maps, mkvc
from SimPEG.EM.FDEM.SrcFDEM import BaseFDEMSrc as FDEMBaseSrc
from SimPEG.EM.Utils import omega
from .Utils.sourceUtils import homo1DModelSource
from .Utils.sourceUtils import analytic1DModelSource

#################
###   Sources ###
#################

class BaseNSEMSrc(FDEMBaseSrc):
    '''
    Sources for the NSEM problem.
    Use the SimPEG BaseSrc, since the source fields share properties with the transmitters.

    :param float freq: The frequency of the source
    :param list rxList: A list of receivers associated with the source
    '''

    freq = None #: Frequency (float)


    def __init__(self, rxList, freq):

        self.freq = float(freq)
        FDEMBaseSrc.__init__(self, rxList)

# 1D sources
class Planewave_xy_1DhomotD(BaseNSEMSrc):
    """
    NSEM source for both polarizations (x and y) for the total Domain.

    It calculates fields calculated based on conditions on the boundary of the domain.
    """
    def __init__(self, rxList, freq):
        BaseNSEMSrc.__init__(self, rxList, freq)


# Need to implement such that it works for all dims.
# Rename to be more descriptive
class Planewave_xy_1Dprimary(BaseNSEMSrc):
    """
    NSEM planewave source for both polarizations (x and y)
    estimated from a single 1D primary models.


    """
    #def __init__(self, rxList, freq):
    def __init__(self, rxList, freq, sigBG1d, sigBG):    # (mod. Diego em 14 dez 2019)
        # assert mkvc(self.mesh.hz.shape,1) == mkvc(sigma1d.shape,1),'The number of values in the 1D background model does not match the number of vertical cells (hz).'
        self.sigma1d = sigBG1d #None                       # (mod. Diego em 14 dez 2019)
        self.sigmaBG = sigBG                               # (mod. Diego em 14 dez 2019)
        BaseNSEMSrc.__init__(self, rxList, freq)

    # campo primário para octree    <<<<<<<<<(mod. Diego em 14 dez 2019)
    
    def ePrimary(self, problem):
        # Get primary fields for both polarizations
        if self.sigma1d is None:
            # Set the sigma1d as the 1st column in the background model
            if len(problem._sigmaPrimary) == problem.mesh.nC:                
                if problem.mesh.dim == 1:
                    self.sigma1d = problem.mesh.r(
                        problem._sigmaPrimary, 'CC', 'CC', 'M')[:]
                elif problem.mesh.dim == 3:
                    self.sigma1d = problem.mesh.r(
                        problem._sigmaPrimary, 'CC', 'CC', 'M')[0, 0, :]
            # Or as the 1D model that matches the vertical cell number
            else: #len(problem._sigmaPrimary) == problem.mesh.nCz:
                self.sigma1d = problem._sigmaPrimary
        
        if self._ePrimary is None:
            self._ePrimary = homo1DModelSource(
                problem.mesh, self.freq, self.sigma1d)
            
#            self._ePrimary = analytic1DModelSource(problem.mesh, self.freq, self.sigma1d)
            
            
        return self._ePrimary 

# (mod. Diego em 14 dez 2019) --------------------------------------------------
        
        # campo primario para tensormesh
        
#    def ePrimary(self, problem):
#        # Get primary fields for both polarizations
#        if self.sigma1d is None:
#            # Set the sigma1d as the 1st column in the background model
#            if len(problem._sigmaPrimary) == problem.mesh.nC:
#                if problem.mesh.dim == 1:
#                    self.sigma1d = problem.mesh.r(
#                        problem._sigmaPrimary, 'CC', 'CC', 'M')[:]
#                elif problem.mesh.dim == 3:
#                    self.sigma1d = problem.mesh.r(
#                        problem._sigmaPrimary, 'CC', 'CC', 'M')[0, 0, :]
#            # Or as the 1D model that matches the vertical cell number
#            elif len(problem._sigmaPrimary) == problem.mesh.nCz:
#                self.sigma1d = problem._sigmaPrimary
#
#        if self._ePrimary is None:
#            self._ePrimary = homo1DModelSource(
#                problem.mesh, self.freq, self.sigma1d)
#        return self._ePrimary
  
    
   
        
    

#------------------------------------------------------------------------------
    
    
    

    def bPrimary(self, problem):
        # Project ePrimary to bPrimary
        # Satisfies the primary(background) field conditions
        
        if problem.mesh.dim == 1:
            C = problem.mesh.nodalGrad            
        elif problem.mesh.dim == 3:
            C = problem.mesh.edgeCurl
        bBG_bp = (- C * self.ePrimary(problem)) * (1 / (1j * omega(self.freq)))
        
        return bBG_bp
    
    
    
  

    def S_e(self, problem):
        """
        Get the electrical field source
        """
        e_p = self.ePrimary(problem)
        #Map_sigma_p = Maps.SurjectVertical1D(problem.mesh)             # (mod. Diego em 14 dez 2019)
        sigma_p = self.sigmaBG #Map_sigma_p._transform(self.sigma1d)    # (mod. Diego em 14 dez 2019)
        # Make mass matrix
        # Note: M(sig) - M(sig_p) = M(sig - sig_p)
        # Need to deal with the edge/face discrepencies between 1d/2d/3d
        if problem.mesh.dim == 1:
            Mesigma = problem.mesh.getFaceInnerProduct(problem.sigma)
            Mesigma_p = problem.mesh.getFaceInnerProduct(sigma_p)
        if problem.mesh.dim == 2:
            pass
        if problem.mesh.dim == 3:
            Mesigma = problem.MeSigma
            Mesigma_p = problem.mesh.getEdgeInnerProduct(sigma_p)
        return (Mesigma - Mesigma_p) * e_p

    def S_eDeriv(self, problem, v, adjoint=False):
        """
        The derivative of S_e with respect to
        """

        return self.S_eDeriv_m(problem, v, adjoint)

    def S_eDeriv_m(self, problem, v, adjoint=False):
        '''
        Get the derivative of S_e wrt to sigma (m)
        '''
        # Need to deal with
        if problem.mesh.dim == 1:
            # Need to use the faceInnerProduct
            ePri = self.ePrimary(problem)[:, 1]
            MsigmaDeriv = (
                problem.mesh.getFaceInnerProductDeriv(problem.sigma)(ePri) *
                problem.sigmaDeriv)
            # MsigmaDeriv = ( MsigmaDeriv * MsigmaDeriv.T)**2

            if adjoint:
                #
                return MsigmaDeriv.T * v
            else:
                # v should be nC size
                return MsigmaDeriv * v
        if problem.mesh.dim == 2:
            raise NotImplementedError('The NSEM 2D problem is not implemented')
        if problem.mesh.dim == 3:
            # Need to take the derivative of both u_px and u_py
            # And stack them to be of the correct size
            e_p = self.ePrimary(problem)
            if adjoint:
                return (
                    problem.MeSigmaDeriv(
                        e_p[:, 0], v[:int(v.shape[0]/2)], adjoint
                    ) +
                    problem.MeSigmaDeriv(
                        e_p[:, 1], v[int(v.shape[0]/2):], adjoint
                    )
                )
                # return sp.hstack((
                #     problem.MeSigmaDeriv(e_p[:, 0]).T,
                #     problem.MeSigmaDeriv(e_p[:, 1]).T)) * v
            else:
                return np.hstack((
                    mkvc(problem.MeSigmaDeriv(e_p[:, 0], v, adjoint), 2),
                    mkvc(problem.MeSigmaDeriv(e_p[:, 1], v, adjoint), 2)))


class Planewave_xy_3Dprimary(BaseNSEMSrc):
    """
    NSEM source for both polarizations (x and y) given a 3D primary model.
    It assigns fields calculated from the 1D model
    as fields in the full space of the problem.
    """
    def __init__(self, rxList, freq):
        # assert mkvc(self.mesh.hz.shape,1) == mkvc(sigma1d.shape,1),'The number of values in the 1D background model does not match the number of vertical cells (hz).'
        self.sigmaPrimary = None
        BaseNSEMSrc.__init__(self, rxList, freq)
        # Hidden property of the ePrimary
        self._ePrimary = None

    def ePrimary(self, problem):
        # Get primary fields for both polarizations
        self.sigmaPrimary = problem._sigmaPrimary

        if self._ePrimary is None:
            self._ePrimary = homo3DModelSource(                    
                problem.mesh, self.sigmaPrimary, self.freq)
        return self._ePrimary

    def bPrimary(self, problem):
        # Project ePrimary to bPrimary
        # Satisfies the primary(background) field conditions
        if problem.mesh.dim == 1:
            C = problem.mesh.nodalGrad
        elif problem.mesh.dim == 3:
            C = problem.mesh.edgeCurl
        bBG_bp = (
            (- C * self.ePrimary(problem)) *
            (1 / (1j * omega(self.freq))))
        return bBG_bp

    def S_e(self, problem):
        """
        Get the electrical field source
        """
        e_p = self.ePrimary(problem)
        Map_sigma_p = Maps.SurjectVertical1D(problem.mesh)
        sigma_p = Map_sigma_p._transform(self.sigma1d)
        # Make mass matrix
        # Note: M(sig) - M(sig_p) = M(sig - sig_p)
        # Need to deal with the edge/face discrepencies between 1d/2d/3d
        if problem.mesh.dim == 1:
            Mesigma = problem.mesh.getFaceInnerProduct(problem.sigma)
            Mesigma_p = problem.mesh.getFaceInnerProduct(sigma_p)
        if problem.mesh.dim == 2:
            pass
        if problem.mesh.dim == 3:
            Mesigma = problem.MeSigma
            Mesigma_p = problem.mesh.getEdgeInnerProduct(sigma_p)
        return (Mesigma - Mesigma_p) * e_p

    def S_eDeriv_m(self, problem, v, adjoint=False):
        '''
        Get the derivative of S_e wrt to sigma (m)
        '''
        # Need to deal with
        if problem.mesh.dim == 1:
            # Need to use the faceInnerProduct
            MsigmaDeriv = (
                problem.mesh.getFaceInnerProductDeriv(
                    problem.sigma)(self.ePrimary(problem)[:, 1]) *
                problem.sigmaDeriv)
            # MsigmaDeriv = ( MsigmaDeriv * MsigmaDeriv.T)**2
        if problem.mesh.dim == 2:
            pass
        if problem.mesh.dim == 3:
            # Need to take the derivative of both u_px and u_py
            ePri = self.ePrimary(problem)
            if adjoint:
                return (
                    problem.MeSigmaDeriv(
                        ePri[:, 0], v[:int(v.shape[0]/2)], adjoint
                    ) +
                    problem.MeSigmaDeriv(
                        ePri[:, 1], v[int(v.shape[0]/2):], adjoint
                    )
                )
                # return sp.hstack((
                #     problem.MeSigmaDeriv(ePri[:, 0]).T,
                #     problem.MeSigmaDeriv(ePri[:, 1]).T)) * v
            else:
                return np.hstack((
                    mkvc(problem.MeSigmaDeriv(ePri[:, 0], v, adjoint), 2),
                    mkvc(problem.MeSigmaDeriv(ePri[:, 1], v, adjoint), 2)
                ))
        if adjoint:
            #
            return MsigmaDeriv.T * v
        else:
            # v should be nC size
            return MsigmaDeriv * v



import numpy as np
from SimPEG import Utils
from scipy.constants import mu_0, epsilon_0
from SimPEG.EM.Utils.EMUtils import k

def getKc(freq,sigma,a,b,mu=mu_0,eps=epsilon_0):
    a = float(a)
    b = float(b)
    # return 1./(2*np.pi) * np.sqrt(b / a) * np.exp(-1j*k(freq,sigma,mu,eps)*(b-a))
    return np.sqrt(b / a) * np.exp(-1j*k(freq,sigma,mu,eps)*(b-a))

def _r2(xyz):
    return np.sum(xyz**2,1)

def _getCasingHertzMagDipole(srcloc,obsloc,freq,sigma,a,b,mu=mu_0*np.ones(3),eps=epsilon_0,moment=1.):
    Kc1 = getKc(freq,sigma[1],a,b,mu[1],eps)

    nobs = obsloc.shape[0]
    dxyz = obsloc - np.c_[np.ones(nobs)]*np.r_[srcloc]

    r2 = _r2(dxyz[:,:2])
    sqrtr2z2 = np.sqrt(r2 + dxyz[:,2]**2)
    k2 = k(freq,sigma[2],mu[2],eps)

    return Kc1 * moment / (4.*np.pi) *np.exp(-1j*k2*sqrtr2z2) / sqrtr2z2


def _getCasingHertzMagDipoleDeriv_r(srcloc,obsloc,freq,sigma,a,b,mu=mu_0*np.ones(3),eps=epsilon_0,moment=1.):
    HertzZ = _getCasingHertzMagDipole(srcloc,obsloc,freq,sigma,a,b,mu,eps,moment)

    nobs = obsloc.shape[0]
    dxyz = obsloc - np.c_[np.ones(nobs)]*np.r_[srcloc]

    r2 = _r2(dxyz[:,:2])
    sqrtr2z2 = np.sqrt(r2 + dxyz[:,2]**2)
    k2 = k(freq,sigma[2],mu[2],eps)

    return -HertzZ * np.sqrt(r2) / sqrtr2z2 * (1j*k2 + 1./ sqrtr2z2)


def _getCasingHertzMagDipoleDeriv_z(srcloc,obsloc,freq,sigma,a,b,mu=mu_0*np.ones(3),eps=epsilon_0,moment=1.):
    HertzZ = _getCasingHertzMagDipole(srcloc,obsloc,freq,sigma,a,b,mu,eps,moment)

    nobs = obsloc.shape[0]
    dxyz = obsloc - np.c_[np.ones(nobs)]*np.r_[srcloc]

    r2z2 = _r2(dxyz)
    sqrtr2z2 = np.sqrt(r2z2)
    k2 = k(freq,sigma[2],mu[2],eps)

    return -HertzZ*dxyz[:,2] /sqrtr2z2 * (1j*k2 + 1./sqrtr2z2)

def _getCasingHertzMagDipole2Deriv_z_r(srcloc,obsloc,freq,sigma,a,b,mu=mu_0*np.ones(3),eps=epsilon_0,moment=1.):
    HertzZ = _getCasingHertzMagDipole(srcloc,obsloc,freq,sigma,a,b,mu,eps,moment)
    dHertzZdr = _getCasingHertzMagDipoleDeriv_r(srcloc,obsloc,freq,sigma,a,b,mu,eps,moment)

    nobs = obsloc.shape[0]
    dxyz = obsloc - np.c_[np.ones(nobs)]*np.r_[srcloc]

    r2 = _r2(dxyz[:,:2])
    r = np.sqrt(r2)
    z = dxyz[:,2]
    sqrtr2z2 = np.sqrt(r2 + z**2)
    k2 = k(freq,sigma[2],mu[2],eps)

    return dHertzZdr*(-z/sqrtr2z2)*(1j*k2+1./sqrtr2z2) + HertzZ*(z*r/sqrtr2z2**3)*(1j*k2 + 2./sqrtr2z2)

def _getCasingHertzMagDipole2Deriv_z_z(srcloc,obsloc,freq,sigma,a,b,mu=mu_0*np.ones(3),eps=epsilon_0,moment=1.):
    HertzZ = _getCasingHertzMagDipole(srcloc,obsloc,freq,sigma,a,b,mu,eps,moment)
    dHertzZdz = _getCasingHertzMagDipoleDeriv_z(srcloc,obsloc,freq,sigma,a,b,mu,eps,moment)

    nobs = obsloc.shape[0]
    dxyz = obsloc - np.c_[np.ones(nobs)]*np.r_[srcloc]

    r2 = _r2(dxyz[:,:2])
    r = np.sqrt(r2)
    z = dxyz[:,2]
    sqrtr2z2 = np.sqrt(r2 + z**2)
    k2 = k(freq,sigma[2],mu[2],eps)

    return (dHertzZdz*z + HertzZ)/sqrtr2z2*(-1j*k2 - 1./sqrtr2z2) + HertzZ*z/sqrtr2z2**3*(1j*k2*z + 2.*z/sqrtr2z2)

def getCasingEphiMagDipole(srcloc,obsloc,freq,sigma,a,b,mu=mu_0*np.ones(3),eps=epsilon_0,moment=1.):
    return 1j * omega(freq) * mu * _getCasingHertzMagDipoleDeriv_r(srcloc,obsloc,freq,sigma,a,b,mu,eps,moment)

def getCasingHrMagDipole(srcloc,obsloc,freq,sigma,a,b,mu=mu_0*np.ones(3),eps=epsilon_0,moment=1.):
    return _getCasingHertzMagDipole2Deriv_z_r(srcloc,obsloc,freq,sigma,a,b,mu,eps,moment)

def getCasingHzMagDipole(srcloc,obsloc,freq,sigma,a,b,mu=mu_0*np.ones(3),eps=epsilon_0,moment=1.):
    d2HertzZdz2 = _getCasingHertzMagDipole2Deriv_z_z(srcloc,obsloc,freq,sigma,a,b,mu,eps,moment)
    k2 = k(freq,sigma[2],mu[2],eps)
    HertzZ = _getCasingHertzMagDipole(srcloc,obsloc,freq,sigma,a,b,mu,eps,moment)
    return d2HertzZdz2 + k2**2 * HertzZ

def getCasingBrMagDipole(srcloc,obsloc,freq,sigma,a,b,mu=mu_0*np.ones(3),eps=epsilon_0,moment=1.):
    return mu_0 * getCasingHrMagDipole(srcloc,obsloc,freq,sigma,a,b,mu,eps,moment)

def getCasingBzMagDipole(srcloc,obsloc,freq,sigma,a,b,mu=mu_0*np.ones(3),eps=epsilon_0,moment=1.):
    return mu_0 * getCasingHzMagDipole(srcloc,obsloc,freq,sigma,a,b,mu,eps,moment)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import matplotlib.pyplot as plt
import numpy as np
import numpy.lib.recfunctions as recFunc
from scipy.constants import mu_0
from scipy import interpolate as sciint

import SimPEG as simpeg
from SimPEG.EM.NSEM.SurveyNSEM import Survey, Data
from SimPEG.EM.NSEM.RxNSEM import (Point_impedance1D,
    Point_impedance3D, Point_tipper3D)
from SimPEG.EM.NSEM.SrcNSEM import Planewave_xy_1Dprimary
from SimPEG.EM.NSEM.Utils import MT1Danalytic, plotDataTypes as pDt



def rotate_data(NSEMdata, rot_angle):
    '''
    Function that rotates clockwise by rotation angle
        (- negative for a counter-clockwise rotation)

    **Required**
    :param SimPEG.EM.NSEM.Data NSEMdata: NSEM data object to process
    :param float rot_angle: Rotation angel in degrees,
        positive for clockwise rotation
    '''
    recData = NSEMdata.toRecArray('Complex')
    impData = rec_to_ndarr(
        recData[['zxx', 'zxy', 'zyx', 'zyy']].copy(), complex)
    # Make the rotation matrix
    # c,s,zxx,zxy,zyx,zyy = sympy.symbols('c,s,zxx,zxy,zyx,zyy')
    # rotM = sympy.Matrix([[c,-s],[s, c]])
    # zM = sympy.Matrix([[zxx,zxy],[zyx,zyy]])
    # rotM*zM*rotM.T
    # [c*(c*zxx - s*zyx) - s*(c*zxy - s*zyy), c*(c*zxy - s*zyy) + s*(c*zxx - s*zyx)],
    # [c*(c*zyx + s*zxx) - s*(c*zyy + s*zxy), c*(c*zyy + s*zxy) + s*(c*zyx + s*zxx)]])
    s = np.sin(-np.deg2rad(rot_angle))
    c = np.cos(-np.deg2rad(rot_angle))
    rotMat = np.array([[c, -s], [s, c]])
    rotData = (
        rotMat.dot(impData.reshape(-1, 2, 2).dot(rotMat.T))
    ).transpose(1, 0, 2).reshape(-1, 4)
    outRec = recData.copy()
    for nr, comp in enumerate(['zxx', 'zxy', 'zyx', 'zyy']):
        outRec[comp] = rotData[:, nr]

    return Data.fromRecArray(outRec)

# Function to get data and data info
def extract_data_info(NSEMdata):
    """
    Simple function that extracts data, frequency and receiver type lists.

    Useful when assigning uncertainties to data based on frequencies and
    receiver types.

    **Required**
    :param SimPEG.EM.NSEM.Data NSEMdata: NSEM data object to process

    """
    dL, freqL, rxTL = [], [], []

    for src in NSEMdata.survey.srcList:
        for rx in src.rxList:
            dL.append(NSEMdata[src, rx])
            freqL.append(np.ones(rx.nD) * src.freq)
            if isinstance(rx, Point_impedance3D):
                rxTL.extend((('z' + rx.orientation + ' ') * rx.nD).split())
            if isinstance(rx, Point_tipper3D):
                rxTL.extend((('t' + rx.orientation + ' ') * rx.nD).split())
    return np.concatenate(dL), np.concatenate(freqL), np.array(rxTL)


def resample_data(NSEMdata, locs='All', freqs='All', rxs='All', verbose=False):
    """
    Function that selects locations from all the receivers in the survey
    (uses the numerator location as a reference). Also gives the option
    of selecting frequencies and receiver.

    **Required**
    :param SimPEG.EM.NSEM.Data NSEMdata: NSEM data object to process

    **Optional**
    :param numpy.ndarray locs: receiver locations to use (default is 'All' locations)
    :param numpy.ndarray freqs: frequencies to use (default is 'All' frequencies))
    :param string rxs: list of receiver sting types to use (default is 'All' types)
        Can be any componation of ['zxx','zxy','zyx','zyy','tzx','tzy']
    """

    # Initiate new objects
    new_srcList = []
    data_list = []
    std_list = []
    floor_list = []

    # Sort out input frequencies
    if locs is 'All':
        locations = NSEMdata._unique_locations()
    elif isinstance(locs, np.ndarray):
        locations = locs
    else:
        raise IOError('Incorrect input type for locs. \n' +
                      'Can be \'All\' or ndarray ')
    # Sort out input frequencies
    if freqs is 'All':
        frequencies = NSEMdata.survey.freqs
    elif isinstance(freqs, np.ndarray):
        frequencies = freqs
    elif isinstance(freqs, list):
        frequencies = np.array(freqs)
    else:
        raise IOError('Incorrect input type for freqs. \n' +
                      'Can be \'All\'; ndarray or a list')
    # Sort out input rxs
    if rxs is 'All':
        rx_comp = True
    elif isinstance(rxs, list):
        rx_comp = []
        for rxT in rxs:
            if'z' in rxT[0]:
                rxtype = Point_impedance3D
            elif 't' in rxT[0]:
                rxtype = Point_tipper3D
            else:
                raise IOError('Unknown rx type string')
            orient = rxT[1:3]
            rx_comp.append((rxtype, orient))

    else:
        raise IOError('Incorrect input type for rxs. \n' +
                      'Can be \'All\' or a list')

    # Filter the data
    for src in NSEMdata.survey.srcList:
        if src.freq in frequencies:
            new_rxList = []
            for rx in src.rxList:
                if rx_comp is True or np.any(
                        [(isinstance(rx, ct) and rx.orientation in co)
                            for (ct, co) in rx_comp]):
                    if len(rx.locs.shape) == 3:
                        ind_loc = np.sum(
                            np.concatenate(
                                [(np.sqrt(np.sum((rx.locs[:, :, 0] - location) ** 2, axis=1)) < 0.1).reshape(-1, 1)
                                 for location in locations],
                                axis=1), axis=1, dtype=bool)
                        new_locs = rx.locs[ind_loc,:,:]
                    else:
                        ind_loc = np.sum(
                            np.concatenate(
                                [(np.sqrt(np.sum((rx.locs[:, :] - location) ** 2, axis=1)) < 0.1).reshape(-1, 1)
                                 for location in locations],
                                axis=1), axis=1, dtype=bool)
                        new_locs = rx.locs[ind_loc, :]
                    new_rx = type(rx)
                    new_rxList.append(
                        new_rx(new_locs, rx.orientation, rx.component))
                    data_list.append(NSEMdata[src, rx][ind_loc])
                    try:
                        std_list.append(
                            NSEMdata.standard_deviation[src, rx][ind_loc])
                        floor_list.append(NSEMdata.floor[src, rx][ind_loc])
                    except Exception as e:
                        if verbose:
                            print('No standard deviation or floor assigned')

            new_src = type(src)
            new_srcList.append(new_src(new_rxList, src.freq))

    survey = Survey(new_srcList)
    if std_list or floor_list:
        return Data(
            survey, np.concatenate(data_list),
            np.concatenate(std_list), np.concatenate(floor_list))
    else:
        return Data(survey, np.concatenate(data_list))

def convert3Dto1Dobject(NSEMdata, rxType3D='yx'):
    """
    Function that converts a 3D NSEMdata of a list of
    1D NSEMdata objects for running 1D inversions for.

    **Required**
    :param SimPEG.EM.NSEM.Data NSEMdata: NSEM data object to process

    **Optional**
    :param string rxType3D: component of the NSEMdata to use
        Can be 'xy', 'yx' or 'det'
    """

    # Find the unique locations
    # Need to find the locations
    recDataTemp = NSEMdata.toRecArray().data.flatten()
    # Check if survey.std has been assigned.
    ## NEED TO: write this...
    # Calculte and add the DET of the tensor to the recArray
    if 'det' in rxType3D:
        Zon = ((recDataTemp['zxxr'] + 1j * recDataTemp['zxxi']) *
               (recDataTemp['zyyr'] + 1j * recDataTemp['zyyi']))
        Zoff = ((recDataTemp['zxyr'] + 1j * recDataTemp['zxyi']) *
                (recDataTemp['zyxr'] + 1j * recDataTemp['zyxi']))
        det = np.sqrt(Zon - Zoff)
        recData = recFunc.append_fields(
            recDataTemp, ['zdetr', 'zdeti'], [det.real, det.imag])
    else:
        recData = recDataTemp

    uniLocs = rec_to_ndarr(np.unique(recData[['x', 'y', 'z']].copy()))
    mtData1DList = []
    if 'xy' in rxType3D:
        corr = -1
        # Shift the data to comply with the quadtrature of the 1d problem
    else:
        corr = 1
    for loc in uniLocs:
        # Make the receiver list
        rx1DList = []
        rx1DList.append(Point_impedance1D(simpeg.mkvc(loc, 2).T, 'real'))
        rx1DList.append(Point_impedance1D(simpeg.mkvc(loc, 2).T, 'imag'))
        # Source list
        locrecData = recData[np.sqrt(
            np.sum((rec_to_ndarr(
                recData[['x', 'y', 'z']].copy()) - loc)**2, axis=1)
        ) < 1e-5]
        dat1DList = []
        src1DList = []
        for freq in locrecData['freq']:
            src1DList.append(Planewave_xy_1Dprimary(rx1DList, freq))
            for comp in ['r', 'i']:
                dat1DList.append(
                    corr *
                    locrecData[rxType3D + comp][locrecData['freq'] == freq])

        # Make the survey
        sur1D = Survey(src1DList)

        # Make the data
        dataVec = np.hstack(dat1DList)
        dat1D = Data(sur1D, dataVec)
        sur1D.dobs = dataVec
        # Need to take NSEMdata.survey.std and split it as well.
        std = 0.05
        sur1D.std = np.abs(sur1D.dobs * std)
        mtData1DList.append(dat1D)

    # Return the the list of data.
    return mtData1DList

### Other utils, that don't take NSEM as an input
def appResPhs(freq, z):
    app_res = ((1. / (8e-7 * np.pi**2)) / freq) * np.abs(z)**2
    app_phs = np.arctan2(z.imag, z.real) * (180 / np.pi)
    return app_res, app_phs

def skindepth(rho, freq):
    ''' Function to calculate the skindepth of EM waves'''
    return np.sqrt((rho * ((1 / (freq * mu_0 * np.pi)))))

def rec_to_ndarr(rec_arr, data_type=float):
    """
    Function to transform a numpy record array to a nd array.
    """
    return rec_arr.copy().view((data_type, len(rec_arr.dtype.names)))

def makeAnalyticSolution(mesh, model, elev, freqs):


    data1D = []
    for freq in freqs:
        anaEd, anaEu, anaHd, anaHu = MT1Danalytic.getEHfields(
            mesh, model, freq, elev)
        anaE = anaEd + anaEu
        anaH = anaHd + anaHu

        anaZ = anaE / anaH
        # Add to the list
        data1D.append((freq, 0, 0, elev, anaZ[0]))
    dataRec = np.array(
        data1D,
        dtype=[
            ('freq', float), ('x', float), ('y', float),
            ('z', float), ('zyx', complex)])
    return dataRec


def plotMT1DModelData(problem, models, symList=None):

    # Setup the figure
    fontSize = 15

    fig = plt.figure(figsize=[9, 7])
    axM = fig.add_axes([0.075, .1, .25, .875])
    axM.set_xlabel('Resistivity [Ohm*m]', fontsize=fontSize)
    axM.set_xlim(1e-1, 1e5)
    axM.set_ylim(-10000, 5000)
    axM.set_ylabel('Depth [km]', fontsize=fontSize)
    axR = fig.add_axes([0.42, .575, .5, .4])
    axR.set_xscale('log')
    axR.set_yscale('log')
    axR.invert_xaxis()
    # axR.set_xlabel('Frequency [Hz]')
    axR.set_ylabel('Apparent resistivity [Ohm m]', fontsize=fontSize)

    axP = fig.add_axes([0.42, .1, .5, .4])
    axP.set_xscale('log')
    axP.invert_xaxis()
    axP.set_ylim(0, 90)
    axP.set_xlabel('Frequency [Hz]', fontsize=fontSize)
    axP.set_ylabel('Apparent phase [deg]', fontsize=fontSize)

    # if not symList:
    #   symList = ['x']*len(models)
    # Loop through the models.
    modelList = [problem.survey.mtrue]
    modelList.extend(models)
    if False:
        modelList = [problem.sigmaMap * mod for mod in modelList]
    for nr, model in enumerate(modelList):
        # Calculate the data
        if nr == 0:
            data1D = problem.dataPair(
                problem.survey, problem.survey.dobs).toRecArray('Complex')
        else:
            data1D = problem.dataPair(
                problem.survey,
                problem.survey.dpred(model)).toRecArray('Complex')
        # Plot the data and the model
        colRat = nr / ((len(modelList) - 1.999) * 1.)
        if colRat > 1.:
            col = 'k'
        else:
            col = plt.cm.seismic(1 - colRat)
        # The model - make the pts to plot
        meshPts = np.concatenate((
            problem.mesh.gridN[0:1],
            np.kron(problem.mesh.gridN[1::], np.ones(2))[:-1]))
        modelPts = np.kron(1. / (problem.sigmaMap * model), np.ones(2,))
        axM.semilogx(modelPts, meshPts, color=col)

        ## Data
        loc = rec_to_ndarr(np.unique(data1D[['x', 'y']]).copy())
        # Appres
        pDt.plotIsoStaImpedance(axR, loc, data1D, 'zyx', 'res', pColor=col)
        # Appphs
        pDt.plotIsoStaImpedance(axP, loc, data1D, 'zyx', 'phs', pColor=col)
        try:
            allData = np.concatenate((
                allData, simpeg.mkvc(data1D['zyx'], 2)), 1)
        except:
            allData = simpeg.mkvc(data1D['zyx'], 2)
    freq = simpeg.mkvc(data1D['freq'], 2)
    res, phs = appResPhs(freq, allData)

    if False:
        stdCol = 'gray'
        axRtw = axR.twinx()
        axRtw.set_ylabel('Std of log10', color=stdCol)
        [(t.set_color(stdCol), t.set_rotation(-45))
            for t in axRtw.get_yticklabels()]
        axPtw = axP.twinx()
        axPtw.set_ylabel('Std ', color=stdCol)
        [t.set_color(stdCol) for t in axPtw.get_yticklabels()]
        axRtw.plot(freq, np.std(np.log10(res), 1), '--', color=stdCol)
        axPtw.plot(freq, np.std(phs, 1), '--', color=stdCol)

    # Fix labels and ticks

    # yMtick = [l/1000 for l in axM.get_yticks().tolist()]
    # axM.set_yticklabels(yMtick)
    [l.set_rotation(90) for l in axM.get_yticklabels()]
    [l.set_rotation(90) for l in axR.get_yticklabels()]
    # [(t.set_color(stdCol), t.set_rotation(-45)) for t in axRtw.get_yticklabels()]
    # [t.set_color(stdCol) for t in axPtw.get_yticklabels()]
    for ax in [axM, axR, axP]:
        ax.xaxis.set_tick_params(labelsize=fontSize)
        ax.yaxis.set_tick_params(labelsize=fontSize)
    return fig

def plotImpAppRes(dataArrays, plotLoc, textStr=[]):
    '''
    Plots amplitude impedance and phase
    '''
    # Make the figure and axes
    fig,axT=plt.subplots(2, 2, sharex=True)
    axes = axT.ravel()
    fig.set_size_inches((13.5, 7.0))
    fig.suptitle(
        '{:s}\nStation at: {:.1f}x ; {:.1f}y'.format(
            textStr, plotLoc[0], plotLoc[1]))
    # Have to deal with axes
    # Set log
    for ax in axes.ravel():
        ax.set_xscale('log')

    axes[0].invert_xaxis()
    axes[0].set_yscale('log')
    axes[2].set_yscale('log')
    # Set labels
    axes[2].set_xlabel('Frequency [Hz]')
    axes[3].set_xlabel('Frequency [Hz]')
    axes[0].set_ylabel('Apperent resistivity [Ohm m]')
    axes[1].set_ylabel('Apperent phase [degrees]')
    axes[1].set_ylim(-180, 180)
    axes[2].set_ylabel('Impedance amplitude [V/A]')
    axes[3].set_ylim(-180, 180)
    axes[3].set_ylabel('Impedance angle [degrees]')


    # Plot the data
    for nr,dataArray in enumerate(dataArrays):
        if nr==1:
            parSym = '*'
        else:
            parSym = 's'
        # app res
        pDt.plotIsoStaImpedance(
            axes[0], plotLoc, dataArray, 'zxy', par='res', pSym=parSym)
        pDt.plotIsoStaImpedance(
            axes[0], plotLoc, dataArray, 'zyx', par='res', pSym=parSym)
        # app phs
        pDt.plotIsoStaImpedance(
            axes[1], plotLoc, dataArray, 'zxy', par='phs', pSym=parSym)
        pDt.plotIsoStaImpedance(
            axes[1], plotLoc, dataArray, 'zyx', par='phs', pSym=parSym)
        # imp abs
        pDt.plotIsoStaImpedance(
            axes[2], plotLoc, dataArray, 'zxx', par='abs', pSym=parSym)
        pDt.plotIsoStaImpedance(
            axes[2], plotLoc, dataArray, 'zxy', par='abs', pSym=parSym)
        pDt.plotIsoStaImpedance(
            axes[2], plotLoc, dataArray, 'zyx', par='abs', pSym=parSym)
        pDt.plotIsoStaImpedance(
            axes[2], plotLoc, dataArray, 'zyy', par='abs', pSym=parSym)
        # imp abs
        pDt.plotIsoStaImpedance(
            axes[3], plotLoc, dataArray, 'zxx', par='phs', pSym=parSym)
        pDt.plotIsoStaImpedance(
            axes[3], plotLoc, dataArray, 'zxy', par='phs', pSym=parSym)
        pDt.plotIsoStaImpedance(
            axes[3], plotLoc, dataArray, 'zyx', par='phs', pSym=parSym)
        pDt.plotIsoStaImpedance(
            axes[3], plotLoc, dataArray, 'zyy', par='phs', pSym=parSym)

    return (fig, axes)


def printTime():
    """
    Small function to print the current time
    """
    import time
    print(time.strftime("%a, %d %b %Y %H:%M:%S +0000", time.localtime()))



from .ProblemTDEM import (
    BaseTDEMProblem, Problem3D_b, Problem3D_e, Problem3D_h, Problem3D_j
)
from .FieldsTDEM import (
    FieldsTDEM, Fields3D_b, Fields3D_e, Fields3D_h, Fields3D_j
)
from .SurveyTDEM import Survey
from . import SrcTDEM as Src
from . import RxTDEM as Rx





from __future__ import print_function
from __future__ import absolute_import

import discretize as Mesh
from discretize import Tests

from . import Maps
from . import Models
from . import Problem
from . import Survey
from . import regularization as Regularization
from . import DataMisfit
from . import InvProblem
from . import Optimization
from . import Directives
from . import Inversion
from . import Tests

from . import Utils
from .Utils import mkvc
from .Utils import Versions
from .Utils.SolverUtils import (
    _checkAccuracy, SolverWrapD, SolverWrapI,
    Solver, SolverCG, SolverDiag, SolverLU, SolverBiCG,
)
__version__   = '0.12.0'
__author__    = 'SimPEG Team'
__license__   = 'MIT'
__copyright__ = '2013 - 2019, SimPEG Team, http://simpeg.xyz'



from __future__ import print_function
import numpy as np
import properties

from . import Utils
from . import Survey
from . import ObjectiveFunction


class BaseDataMisfit(ObjectiveFunction.L2ObjectiveFunction):
    """
    BaseDataMisfit

        .. note::

            You should inherit from this class to create your own data misfit
            term.
    """

    debug   = False  #: Print debugging information
    counter = None  #: Set this to a SimPEG.Utils.Counter() if you want to count things

    _hasFields = True  #: Data Misfits take fields, handy to store them

    def __init__(self, survey, **kwargs):
        assert survey.ispaired, 'The survey must be paired to a problem.'
        if isinstance(survey, Survey.BaseSurvey):
            self.survey = survey
            self.prob   = survey.prob
        super(BaseDataMisfit, self).__init__(**kwargs)

    @property
    def nP(self):
        if self._mapping is not None:
            return self.mapping.nP
        elif self.prob.model is not None:
            return len(self.prob.model)
        else:
            return '*'

    @property
    def Wd(self):
        raise AttributeError(
            'The `Wd` property been depreciated, please use: `W` instead'
        )


class l2_DataMisfit(BaseDataMisfit):
    """

    The data misfit with an l_2 norm:

    .. math::

        \mu_\\text{data} = {1\over 2}\left|
        \mathbf{W}_d (\mathbf{d}_\\text{pred} -
        \mathbf{d}_\\text{obs}) \\right|_2^2

    """

    std = None  #: default standard deviation if not provided by survey
    eps = None  #: default floor
    eps_factor = 1e-5  #: factor to multiply by the norm of the data to create floor

    def __init__(self, survey, **kwargs):
        BaseDataMisfit.__init__(self, survey, **kwargs)

        if self.std is None:
            if getattr(self.survey, 'std', None) is None:
                print(
                    'SimPEG.DataMisfit.l2_DataMisfit assigning default std '
                    'of 5%'
                )
                self.std = 0.05
            else:
                self.std = self.survey.std

        if self.eps is None:
            if getattr(self.survey, 'eps', None) is None:
                print(
                    'SimPEG.DataMisfit.l2_DataMisfit assigning default eps '
                    'of 1e-5 * ||dobs||'
                )
                self.eps = (
                    np.linalg.norm(Utils.mkvc(survey.dobs), 2)*self.eps_factor
                )  # default
            else:
                self.eps = self.survey.eps

    @property
    def W(self):
        """W

            The data weighting matrix.

            The default is based on the norm of the data plus a noise floor.

            :rtype: scipy.sparse.csr_matrix
            :return: W

        """

        if getattr(self, '_W', None) is None:
            survey = self.survey
            self._W = Utils.sdiag(1/(abs(survey.dobs)*self.std+self.eps))

        return self._W

    @W.setter
    def W(self, value):
        if len(value.shape) < 2:
            value = Utils.sdiag(value)
        assert value.shape == (self.survey.nD, self.survey.nD), (
            'W must have shape ({nD},{nD}), not ({val0}, val{1})'.format(
                nD=self.survey.nD, val0=value.shape[0], val1=value.shape[1]
            )
        )
        self._W = value

    @Utils.timeIt
    def __call__(self, m, f=None):
        "__call__(m, f=None)"
        if f is None:
            f = self.prob.fields(m)
        R = self.W * self.survey.residual(m, f)
        return 0.5*np.vdot(R, R)

    @Utils.timeIt
    def deriv(self, m, f=None):
        """
        deriv(m, f=None)

        Derivative of the data misfit

        .. math::

            \mathbf{J}^{\top} \mathbf{W}^{\top} \mathbf{W}
            (\mathbf{d} - \mathbf{d}^{obs})

        :param numpy.ndarray m: model
        :param SimPEG.Fields.Fields f: fields object
        """
        if f is None:
            f = self.prob.fields(m)
        return self.prob.Jtvec(
            m, self.W.T * (self.W * self.survey.residual(m, f=f)), f=f
        )

    @Utils.timeIt
    def deriv2(self, m, v, f=None):
        """
        deriv2(m, v, f=None)

        .. math::

            \mathbf{J}^{\top} \mathbf{W}^{\top} \mathbf{W} \mathbf{J}

        :param numpy.ndarray m: model
        :param numpy.ndarray v: vector
        :param SimPEG.Fields.Fields f: fields object
        """
        if f is None:
            f = self.prob.fields(m)
        return self.prob.Jtvec_approx(
            m, self.W * (self.W * self.prob.Jvec_approx(m, v, f=f)), f=f
        )






from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import matplotlib.pyplot as plt
import numpy as np

import properties
from SimPEG.EM.Utils.EMUtils import mu_0, omega

# Define the default component dictionaries
DEFAULT_COMP_DICT = {
    'xx':{'color': 'green', 'label': 'Imp_xx', 'marker': '_', 'ls': 'None'},
    'xy':{'color': 'blue', 'label': 'Imp_xy', 'marker': '_', 'ls': 'None'},
    'yx':{'color': 'red', 'label': 'Imp_yx', 'marker': '_', 'ls': 'None'},
    'yy':{'color': 'yellow', 'label': 'Imp_yy', 'marker': '_', 'ls': 'None'},
    'zx':{'color': 'brown', 'label': 'Tip_zx', 'marker': '_', 'ls': 'None'},
    'zy':{'color': 'purple', 'label': 'Tip_zy', 'marker': '_', 'ls': 'None'},
}

# Define some hidden attributes
_imp_comps = ['xx', 'xy', 'yx', 'yy']

def _validate_kwargs(input_dict, compare_dict):
    """
    Function to deal with keyword arguments.

    :param input_dict: matplotlib kwargs dictionary with custom arguments
    :type input_dict: :class: `dict`
    :param compare_dict: matplotlib kwargs of default to use arguments
    :type compare_dict: :class: `dict`
    """
    # Set the default plot kwargs
    for key, val in compare_dict.items():
        # Make sure they aren't already assigned
        if input_dict is None:
            input_dict = compare_dict.copy()
        else:
            if key not in input_dict:
                input_dict[key] = val
    # Return
    return input_dict


class BaseDataNSEMPlots(properties.HasProperties):
    """
    A class container of matplotlib panels for plotting
    NSEM data.

    """
    fig = properties.Instance(
        'Figure plotting',
        plt.Figure,
        required=False)
    axes = properties.List(
        'List of plot axes',
        properties.Instance('Axes to plot the on', plt.Axes),
        required=False)

    def setup(self):
        """
        Setup up the plot window.

        Should populate the
            self.fig and self.axes properties
        """
        raise NotImplementedError('Is required in subclasses of {}'.format(self.__class__))

    def draw(self):
        raise NotImplementedError('Is required in subclasses of {}'.format(self.__class__))


    def clear_axes(self):
        """
        Function to clear all of the axes
        """
        for ax in self.axes:
            while len(ax.lines) > 0:
                for line in ax.lines:
                    ax.lines.remove(line)

            while len(ax.collections) > 0:
                for item in ax.collections:
                    item.remove()


class TipperAmplitudeStationPlot(BaseDataNSEMPlots):
    """
    Class for setting up 2 axes figure with:
        tipper amplitudes | tipper phase
        setup.
    """

    def __init__(self):
        super(TipperAmplitudeStationPlot, self).__init__()


    def setup(self):
        """
        Setup a station data plot figure.
        """
        self.fig, axes_temp = plt.subplots(1, 2, sharex=True)
        self.axes = axes_temp.ravel().tolist()
        self.fig.set_size_inches((13.5, 4.0))

        for ax in self.axes:
            ax.set_xscale('log')

        self.axes[0].invert_xaxis()
        self.axes[0].set_yscale('log')
        # Set labels
        self.axes[0].set_xlabel('Frequency [Hz]')
        self.axes[1].set_xlabel('Frequency [Hz]')
        self.axes[0].set_ylabel('Tipper amplitude [V/A]')
        self.axes[1].set_ylim(-180, 180)
        self.axes[1].set_ylabel('Tipper angle [degrees]')

    def draw(self, data_list, location):
        """
        Function to draw on the axes

        :param data_list: List of NSEM data objects to plot.
            Has to be of length >= 1. First item is treat as a
            observed data (Hast to have standard_deviation and floor)
            assigned) and the others are plotted on top.
        :param location: Location of the station to plot
        """

        axes = self.axes

        # Set keyword arguments
        st_kwargs = {'marker': '_', 'ls': 'None'}
        eb_kwargs = {'ls': 'None'}
        # Pop the data from the list
        data = data_list[0]

        # Apparent resistivity
        data.plot_tip_amp(location, ['zx', 'zy'],
                          ax=axes[0], errorbars=True)

        # Apparent phase
        data.plot_app_phs(location, ['zx', 'zy'],
                          ax=axes[1], errorbars=True)

        # Plot the additional data
        for other_data in data_list[1::]:
            # Need add symbol generation
            dd_kwargs = {'zx': {'marker': '.', 'ls': '--'},
                         'zy': {'marker': '.', 'ls': '--'}}

            # Apparent resistivity
            other_data.plot_tip_amp(location, ['zx', 'zy'],
                                    ax=axes[0], errorbars=False,
                                    comp_plot_dict=dd_kwargs)

            # Apparent phase
            other_data.plot_app_phs(location, ['zx', 'zy'],
                                    ax=axes[1], errorbars=False,
                                    comp_plot_dict=dd_kwargs)


class ApperentResPhsStationPlot(BaseDataNSEMPlots):
    """
    Class for setting up 4 axes figure with:
            apparent resistivity | phase
            --------------------------------
            impedance amplitudes | impedance phase
        setup.
    """

    def __init__(self):
        super(ApperentResPhsStationPlot, self).__init__()


    def setup(self):
        """
        Setup a station data plot figure.
        """
        self.fig, axes_temp = plt.subplots(2, 2, sharex=True)
        self.axes = axes_temp.ravel().tolist()
        self.fig.set_size_inches((13.5, 7.0))

        # Have to deal with axes
        # Set log

        for ax in self.axes:
            ax.set_xscale('log')

        self.axes[0].invert_xaxis()
        self.axes[0].set_yscale('log')
        self.axes[2].set_yscale('log')
        # Set labels
        self.axes[2].set_xlabel('Frequency [Hz]')
        self.axes[3].set_xlabel('Frequency [Hz]')
        self.axes[0].set_ylabel('Apperent resistivity [Ohm m]')
        self.axes[1].set_ylabel('Apperent phase [degrees]')
        self.axes[1].set_ylim(-180, 180)
        self.axes[2].set_ylabel('Impedance amplitude [V/A]')
        self.axes[3].set_ylim(-180, 180)
        self.axes[3].set_ylabel('Impedance angle [degrees]')

    def draw(self, data_list, location):
        """
        Function to draw on the axes

        :param data_list: List of NSEM data objects to plot.
            Has to be of length >= 1. First item is treat as a
            observed data (Hast to have standard_deviation and floor)
            assigned) and the others are plotted on top.
        :param location: Location of the station to plot
        """

        axes = self.axes

        # Set keyword arguments
        st_kwargs = {'marker':'_', 'ls':'None'}
        eb_kwargs = {'ls':'None'}
        # Pop the data from the list
        data = data_list[0]

        # Apparent resistivity
        data.plot_app_res(location, ['xy', 'yx'],
                          ax=axes[0], errorbars=True)

        # Apparent phase
        data.plot_app_phs(location, ['xy', 'yx'],
                          ax=axes[1], errorbars=True)

        # Impedamce amplitude
        data.plot_imp_amp(location, ['xx', 'xy', 'yx', 'yy'],
                          ax=axes[2], errorbars=True)

        # Impedance phase
        data.plot_app_phs(location, ['xx', 'xy', 'yx', 'yy'],
                          ax=axes[3], errorbars=True)


        # Plot the additional data
        for other_data in data_list[1::]:
            # Need add symbol generation
            dd_kwargs = {'xx': {'marker': '.', 'ls': '--'},
                         'xy': {'marker': '.', 'ls': '--'},
                         'yx': {'marker': '.', 'ls': '--'},
                         'yy': {'marker': '.', 'ls': '--'}}

            # Apparent resistivity
            other_data.plot_app_res(location, ['xy', 'yx'],
                                    ax=axes[0], errorbars=False,
                                    comp_plot_dict=dd_kwargs)

            # Apparent phase
            other_data.plot_app_phs(location, ['xy', 'yx'],
                                    ax=axes[1], errorbars=False,
                                    comp_plot_dict=dd_kwargs)

            # Impedamce amplitude
            other_data.plot_imp_amp(location,
                                    ['xx', 'xy', 'yx', 'yy'],
                                    ax=axes[2], errorbars=False,
                                    comp_plot_dict=dd_kwargs)

            # Impedance phase
            other_data.plot_app_phs(location,
                                    ['xx', 'xy', 'yx', 'yy'],
                                    ax=axes[3], errorbars=False,
                                    comp_plot_dict=dd_kwargs)


class DataNSEMPlotMethods(object):
    """
    Class container for properties and methods for
    plotting of NSEM data.

    """

    def __init__(self):
        """
        Just for reference
        """
        pass

    def plot_app_res(
        self, location,
        components=['xy', 'yx'], ax=None, errorbars=False,
            comp_plot_dict=DEFAULT_COMP_DICT):
        """
        Plot apperent resistivity curves at a given location

        :param location: Location of the data point
        :type location: :class:`axes <matplotlib.axes>`
        :param components: List of the components to plot.
            Default = ['xy','yx']
        :type components: list
        :param ax: The ax object to add the,  , Default: None
        :type ax: :class:`axes <matplotlib.axes>`
        :param errorbars: Controls if errorbars are plotted
            Default = True
        :type errorbars: boolean
        :param comp_plot_dict: Dictionary with additional kwargs
            for matplotlib.plot
        :type comp_plot_dict: dict

        """
        if ax is None:
            fig, ax = plt.subplots(1, 1)
            ax.invert_xaxis()
            ax.set_xscale('log')
            ax.set_yscale('log')
            ax.set_xlabel('Frequency [Hz]')
            ax.set_ylabel('Apperent resistivity [Ohm m]')
        else:
            fig = ax.get_figure()

        for comp in components:
            st_kwargs = _validate_kwargs(
                comp_plot_dict[comp], DEFAULT_COMP_DICT[comp])

            self.station_component(
                location, comp, 'app_res',
                ax=ax, **st_kwargs)
            if errorbars:
                eb_kwargs = _validate_kwargs(
                    comp_plot_dict[comp], DEFAULT_COMP_DICT[comp])

                self.station_errorbars(
                    location, comp, 'app_res',
                    ax=ax, **eb_kwargs)

        return ax

    def plot_app_phs(
        self, location,
        components=['xy', 'yx'], ax=None, errorbars=False,
            comp_plot_dict=DEFAULT_COMP_DICT):
        """
        Plot apperent resistivity curves at a given location

        :param location: Location of the data point
        :type location: :class:`axes <matplotlib.axes>`
        :param components: List of the components to plot.
            Default = ['xy','yx']
        :type components: list
        :param ax: The ax object to add the,  , Default: None
        :type ax: :class:`axes <matplotlib.axes>`
        :param errorbars: Controls if errorbars are plotted
            Default = True
        :type errorbars: boolean
        :param comp_plot_dict: Dictionary with additional kwargs
            for matplotlib.plot settings
        :type comp_plot_dict: dict
        """
        if ax is None:
            fig, ax = plt.subplots(1, 1)
            ax.invert_xaxis()
            ax.set_xscale('log')
            ax.set_xlabel('Frequency [Hz]')
            ax.set_ylabel('Phase angle [Degrees]')
        else:
            fig = ax.get_figure()

        for comp in components:
            st_kwargs = _validate_kwargs(
                comp_plot_dict[comp], DEFAULT_COMP_DICT[comp])

            self.station_component(
                location, comp, 'phase',
                ax=ax, **st_kwargs)
            if errorbars:
                eb_kwargs = _validate_kwargs(
                    comp_plot_dict[comp], DEFAULT_COMP_DICT[comp])

                self.station_errorbars(
                    location, comp, 'phase',
                    ax=ax, **eb_kwargs)

        return ax

    def plot_imp_amp(
        self, location,
        components=['xy', 'yx'], ax=None, errorbars=False,
            comp_plot_dict=DEFAULT_COMP_DICT):
        """
        Plot impedance amplitude curves at a given location

        :param location: Location of the data point
        :type location: :class:`axes <matplotlib.axes>`
        :param components: List of the components to plot.
            Default = ['xy','yx']
        :type components: list
        :param ax: The ax object to add the,  , Default: None
        :type ax: :class:`axes <matplotlib.axes>`
        :param errorbars: Controls if errorbars are plotted
            Default = True
        :type errorbars: boolean
        :param comp_plot_dict: Dictionary with additional kwargs
            for matplotlib.plot
        :type comp_plot_dict: dict
        """
        if ax is None:
            fig, ax = plt.subplots(1, 1)
            ax.invert_xaxis()
            ax.set_xscale('log')
            ax.set_yscale('log')
            ax.set_xlabel('Frequency [Hz]')
            ax.set_ylabel('Impedance amplitude [V/A]')
        else:
            fig = ax.get_figure()

        for comp in components:
            st_kwargs = _validate_kwargs(
                comp_plot_dict[comp], DEFAULT_COMP_DICT[comp])

            self.station_component(
                location, comp, 'amplitude',
                ax=ax, **st_kwargs)
            if errorbars:
                eb_kwargs = _validate_kwargs(
                    comp_plot_dict[comp], DEFAULT_COMP_DICT[comp])

                self.station_errorbars(
                    location, comp, 'amplitude',
                    ax=ax, **eb_kwargs)

        return ax

    def plot_tip_amp(
        self, location,
        components=['zx', 'zy'], ax=None, errorbars=False,
            comp_plot_dict=DEFAULT_COMP_DICT):
        """
        Plot tipper amplitude curves at a given location

        :param location: Location of the data point
        :type location: :class:`axes <matplotlib.axes>`
        :param components: List of the components to plot.
            Default = ['xy','yx']
        :type components: list
        :param ax: The ax object to add the,  , Default: None
        :type ax: :class:`axes <matplotlib.axes>`
        :param errorbars: Controls if errorbars are plotted
            Default = True
        :type errorbars: boolean
        :param comp_plot_dict: Dictionary with additional kwargs
            for matplotlib.plot
        :type comp_plot_dict: dict
        """
        if ax is None:
            fig, ax = plt.subplots(1, 1)
            ax.invert_xaxis()
            ax.set_xscale('log')
            ax.set_yscale('log')
            ax.set_xlabel('Frequency [Hz]')
            ax.set_ylabel('Tipper magnitude [unitless]')
        else:
            fig = ax.get_figure()

        for comp in components:
            st_kwargs = _validate_kwargs(
                comp_plot_dict[comp], DEFAULT_COMP_DICT[comp])

            self.station_component(
                location, comp, 'amplitude',
                ax=ax, **st_kwargs)
            if errorbars:
                eb_kwargs = _validate_kwargs(
                    comp_plot_dict[comp], DEFAULT_COMP_DICT[comp])

                self.station_errorbars(
                    location, comp, 'amplitude',
                    ax=ax, **eb_kwargs)

        return ax

    def map_data_locations(
        self, ax=None,
            **plot_kwargs):
        """
        Function that plots all receiver locations of the data
            (all discreate data locations).

        :param ax: The ax object for mapping to. Default: None
        :type ax: :class:`axes <matplotlib.axes>`

        """
        # Default plot dict
        default_dict = {
            'marker': '+',
            'c': 'k',
            'ms': 10,
            'ls': 'None',
            'zorder': 4
        }

        # Set the default plot kwargs
        for key, val in default_dict.items():
            # Make sure they aren't already assigned
            if key not in plot_kwargs:
                plot_kwargs[key] = val
        # Get unique locations
        unique_locations = _unique_rows(np.concatenate(
            [rx.locs for src in self.survey.srcList for rx in src.rxList])
        )
        # Make the figure and the axes
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        else:
            fig = ax.get_figure()

        # Plot the locations
        ax.plot(unique_locations[:, 0], unique_locations[:, 1], **plot_kwargs)

        return (fig, ax)

    def station_component(
        self, location, orientation, component,
            ax=None, **plot_kwargs):
        """

        :param numpy.ndarray location: Coordnaties of the station to plot
        :param str orientation: The orientation of the data
        :param str component: The data component to plot
        :param matplotlib.axes ax (optional):
        :param matplotlib.lines.Line2D keyword_arguments plot_kwargs)

        """

        # Sort the axes
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        else:
            fig = ax.get_figure()

        # Plot the data
        freqs, plot_data = _get_plot_data(self,
                                          location, orientation,
                                          component)
        # Plot
        plot_obj = ax.plot(freqs, plot_data, **plot_kwargs)


        return (fig, ax, plot_obj)

    def station_errorbars(
        self, location, orientation, component,
            ax=None, **plot_kwargs):
        """

        :param numpy.ndarray location: Coordnaties of the station to plot
        :param str orientation: The orientation of the data
        :param str component: The data component to plot
        :param matplotlib.axes ax (optional):
        :param matplotlib.lines.Line2D keyword_arguments plot_kwargs)

        """

        # Sort the axes
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        else:
            fig = ax.get_figure()

        # Plot the data
        freqs, plot_data, errorbars = _get_station_data(
            self, location, orientation, component, plot_error=True)
        plot_obj = ax.errorbar(freqs, plot_data, yerr=errorbars,
                               **plot_kwargs)
        return (fig, ax, plot_obj)

    def frequency_map(
        self, frequency, orientation, component,
            ax=None, plot_error=True, **plot_kwargs):
        """
        Function to generate a iso-frequency map

        :param numpy.ndarray frequency: Frequency to be mapped
        :param str orientation: The orientation of the data
        :param str component: The data component to plot
        :param matplotlib.axes ax (optional):
        :param matplotlib.lines.Line2D keyword_arguments plot_kwargs)

        """

        # Sort the axes
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        else:
            fig = ax.get_figure()

        # Plot the data
        locs, plot_data, errorbars = _get_map_data(
            self, frequency, orientation, component,plot_error)
        plot_obj = plt.tricontourf(locs, plot_data,
                               **plot_kwargs)
        return (fig, ax, plot_obj)

# Hidden utils functions
def _get_map_data(
    data, frequency, orientation, component, plot_error=False):
    """
    Function for getting frequency map data
    """

    # Get the components
    if component in ['app_res', 'phase', 'amplitude']:
        real_tuple = _extract_frequency_data(data, frequency, orientation,
                                            'real', plot_error)
        imag_tuple = _extract_frequency_data(data, frequency, orientation,
                                            'imag', plot_error)
        if plot_error:
            freqs, real_data, real_std, real_floor = real_tuple
            freqs, imag_data, imag_std, imag_floor = imag_tuple
            # Add up the uncertainties
            real_uncert = real_std * np.abs(real_data) + real_floor
            imag_uncert = imag_std * np.abs(imag_data) + imag_floor
        else:
            freqs, real_data = real_tuple
            freqs, imag_data = imag_tuple

        if 'app_res' in component:
            comp_data = real_data + 1j * imag_data
            plot_data = (1. / (mu_0 * omega(freqs))) * np.abs(comp_data) ** 2
            if plot_error:
                res_uncert = (
                    (2. / (mu_0 * omega(freqs))) *
                    (real_data * real_uncert + imag_data * imag_uncert)
                )
                errorbars = [res_uncert, res_uncert]
        elif 'phase' in component:
            plot_data = np.arctan2(imag_data, real_data) * (180. / np.pi)
            if plot_error:
                phs_uncert = (
                    (1. / (real_data ** 2 + imag_data ** 2)) *
                    ((real_data * real_uncert - imag_data * imag_uncert))
                ) * (180. / np.pi)
                # Scale back the errorbars
                errorbars = [phs_uncert, phs_uncert]
        elif 'amplitude' in component:
            comp_data = real_data + 1j * imag_data
            plot_data = np.abs(comp_data)
            if plot_error:
                amp_uncert = ((1. / plot_data) *
                              ((np.abs(real_data) * real_uncert) +
                              (np.abs(imag_data) * imag_uncert))
                              )
                errorbars = [amp_uncert, amp_uncert] #[low_unsert, up_unsert]
    else:

        if plot_error:
            freqs, plot_data, std_data, floor_data = _extract_frequency_data(
                data, frequency, orientation, component, return_uncert=error)
            attr_uncert = std_data * np.abs(plot_data) + floor_data
            errorbars = [attr_uncert, attr_uncert]
        else:
            freqs, plot_data = _extract_frequency_data(
                data, frequency, orientation, component, return_uncert=False)
    return (freqs, plot_data, errorbars)

def _get_station_data(
    data, location, orientation, component, plot_error=False):

    # Get the components
    if component in ['app_res', 'phase', 'amplitude']:
        real_tuple = _extract_location_data(data, location, orientation,
                                            'real', plot_error)
        imag_tuple = _extract_location_data(data, location, orientation,
                                            'imag', plot_error)

        if plot_error:
            freqs, real_data, real_std, real_floor = real_tuple
            freqs, imag_data, imag_std, imag_floor = imag_tuple
            # Add up the uncertainties
            real_uncert = real_std * np.abs(real_data) + real_floor
            imag_uncert = imag_std * np.abs(imag_data) + imag_floor
        else:
            freqs, real_data = real_tuple
            freqs, imag_data = imag_tuple

        if 'app_res' in component:
            comp_data = real_data + 1j * imag_data
            plot_data = (1. / (mu_0 * omega(freqs))) * np.abs(comp_data) ** 2
            if plot_error:
                res_uncert = (
                    (2. / (mu_0 * omega(freqs))) *
                    (real_data * real_uncert + imag_data * imag_uncert)
                )
                errorbars = [res_uncert, res_uncert]
        elif 'phase' in component:
            plot_data = np.arctan2(imag_data, real_data) * (180. / np.pi)
            if plot_error:
                phs_uncert = (
                    (1. / (real_data ** 2 + imag_data ** 2)) *
                    ((real_data * real_uncert - imag_data * imag_uncert))
                ) * (180. / np.pi)
                # Scale back the errorbars
                errorbars = [phs_uncert, phs_uncert]
        elif 'amplitude' in component:
            comp_data = real_data + 1j * imag_data
            plot_data = np.abs(comp_data)
            if plot_error:
                amp_uncert = ((1. / plot_data) *
                              ((np.abs(real_data) * real_uncert) +
                              (np.abs(imag_data) * imag_uncert))
                              )
                errorbars = [amp_uncert, amp_uncert] #[low_unsert, up_unsert]
    else:
        if plot_error:
            freqs, plot_data, std_data, floor_data = _extract_location_data(
                data, location, orientation, component, return_uncert=plot_error)
            attr_uncert = std_data * np.abs(plot_data) + floor_data
            errorbars = [attr_uncert, attr_uncert]
        else:
            freqs, plot_data = _extract_location_data(
                data, location, orientation, component, return_uncert=plot_error)
    if plot_error:
        return (freqs, plot_data, errorbars)
    else:
        return (freqs, plot_data)

def _get_plot_data(data, location, orientation, component):

    if 'app_res' in component:
        freqs, dat_r = _extract_location_data(
            data, location, orientation, 'real')
        freqs, dat_i = _extract_location_data(
            data, location, orientation, 'imag')
        dat = dat_r + 1j * dat_i
        plot_data = 1. / (mu_0 * omega(freqs)) * np.abs(dat) ** 2
    elif 'phase' in component:
        freqs, dat_r = _extract_location_data(
            data, location, orientation, 'real')
        freqs, dat_i = _extract_location_data(
            data, location, orientation, 'imag')
        plot_data = np.arctan2(dat_i, dat_r) * (180. / np.pi)
    elif 'amplitude' in component:
        freqs, dat_r = _extract_location_data(
            data, location, orientation, 'real')
        freqs, dat_i = _extract_location_data(
            data, location, orientation, 'imag')
        dat_complex = dat_r + 1j * dat_i
        plot_data = np.abs(dat_complex)
    else:
        freqs, plot_data = _extract_location_data(
            data, location, orientation, component)
    return (freqs, plot_data)

def _extract_frequency_data(
    data, frequency, orientation, component, return_uncert=False):
    """
    Function to extract data at given frequency
    """

    src = data.survey.getSrcByFreq(frequency)
    rx_list = [rx for rx in src.rxList
               if rx.orientation == orientation and rx.component == component]
    # Check the number of the rx
    if len(rx_list) == 1:
        rx = rx_list[0]
    elif len(rx_list) == 0:
        # Should add a warning that the rx doesn't excist
        if return_uncert:
            # Returning all empty arrays
            return (np.array([]), np.array([]),
                    np.array([]), np.array([]))
        return (np.array([]), np.array([]))
    else:
        # Should be a more specifice Exeption
        raise Exception(
            'To many Receivers of the same type, orientation and component')

    loc_arr = rx.locs
    data_arr = data[src, rx]
    if return_uncert:
        std_arr = data.standard_deviation[src, rx]
        floor_arr = data.floor[src, rx]
    if return_uncert:
        return (loc_arr, data_arr, std_arr, floor_arr)
    return (loc_arr, data_arr)

def _extract_location_data(
    data, location, orientation, component, return_uncert=False):
    """
    Function to extract data at given location
    """
    freq_list = []
    data_list = []
    std_list = []
    floor_list = []
    for src in data.survey.srcList:
        rx_list = [rx for rx in src.rxList
              if rx.orientation == orientation and rx.component == component]
        if len(rx_list) == 0:
            if return_uncert:
                return (np.array([]), np.array([]),
                        np.array([]), np.array([]))
            return (np.array([]), np.array([]))
        else:
            rx = rx_list[0]

        ind_loc = np.sqrt(np.sum((rx.locs[:, :2] - location) ** 2, axis=1)) < 0.1
        if np.any(ind_loc):
            freq_list.append(src.freq)
            data_list.append(data[src, rx][ind_loc])


            if return_uncert:
                std_list.append(data.standard_deviation[src, rx][ind_loc])
                floor_list.append(data.floor[src, rx][ind_loc])
    if return_uncert:
        return (np.array(freq_list), np.concatenate(data_list),
                np.concatenate(std_list), np.concatenate(floor_list))
    return (np.array(freq_list), np.concatenate(data_list))

## Hidden utility functions
# Unique row function-should be moved to utils
def _unique_rows(array):
    """
    Finds and returns unique rows in an array
    """
    array = np.ascontiguousarray(array)
    unique_array = np.unique(array.view([('', array.dtype)] * array.shape[1]))
    return unique_array.view(
        array.dtype).reshape((unique_array.shape[0], array.shape[1]))




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from SimPEG import Utils
from SimPEG.EM.Base import BaseEMProblem
from .SurveyDC import Survey
from .FieldsDC import FieldsDC, Fields_CC, Fields_N
import numpy as np
import scipy as sp
from SimPEG.Utils import Zero
from .BoundaryUtils import getxBCyBC_CC


class BaseDCProblem(BaseEMProblem):
    """
    Base DC Problem
    """
    surveyPair = Survey
    fieldsPair = FieldsDC
    Ainv = None
    storeJ = False
    _Jmatrix = None

    def fields(self, m=None):
        if m is not None:
            self.model = m

        if self.Ainv is not None:
            self.Ainv.clean()

        f = self.fieldsPair(self.mesh, self.survey)
        A = self.getA()
        self.Ainv = self.Solver(A, **self.solverOpts)
        RHS = self.getRHS()
        u = self.Ainv * RHS
        Srcs = self.survey.srcList
        f[Srcs, self._solutionType] = u
        return f

    def getJ(self, m, f=None):
        """
            Generate Full sensitivity matrix
        """
        if self.verbose:
            print("Calculating J and storing")

        if self._Jmatrix is not None:
            return self._Jmatrix
        else:

            self.model = m
            if f is None:
                f = self.fields(m)
            self._Jmatrix = (self._Jtvec(m, v=None, f=f)).T
        return self._Jmatrix

    def Jvec(self, m, v, f=None):
        """
            Compute sensitivity matrix (J) and vector (v) product.
        """
        if self.storeJ:
            J = self.getJ(m, f=f)
            Jv = Utils.mkvc(np.dot(J, v))
            return Jv

        self.model = m

        if f is None:
            f = self.fields(m)

        Jv = []

        for src in self.survey.srcList:
            u_src = f[src, self._solutionType]  # solution vector
            dA_dm_v = self.getADeriv(u_src, v)
            dRHS_dm_v = self.getRHSDeriv(src, v)
            du_dm_v = self.Ainv * (- dA_dm_v + dRHS_dm_v)
            for rx in src.rxList:
                df_dmFun = getattr(f, '_{0!s}Deriv'.format(rx.projField), None)
                df_dm_v = df_dmFun(src, du_dm_v, v, adjoint=False)
                Jv.append(rx.evalDeriv(src, self.mesh, f, df_dm_v))
        return np.hstack(Jv)

    def Jtvec(self, m, v, f=None):
        """
            Compute adjoint sensitivity matrix (J^T) and vector (v) product.

        """
        if self.storeJ:
            J = self.getJ(m, f=f)
            Jtv = Utils.mkvc(np.dot(J.T, v))
            return Jtv

        self.model = m

        if f is None:
            f = self.fields(m)

        return self._Jtvec(m, v=v, f=f)

    def _Jtvec(self, m, v=None, f=None):
        """
            Compute adjoint sensitivity matrix (J^T) and vector (v) product.
            Full J matrix can be computed by inputing v=None
        """

        if v is not None:
            # Ensure v is a data object.
            if not isinstance(v, self.dataPair):
                v = self.dataPair(self.survey, v)
            Jtv = np.zeros(m.size)
        else:
            # This is for forming full sensitivity matrix
            Jtv = np.zeros((self.model.size, self.survey.nD), order='F')
            istrt = int(0)
            iend = int(0)

        for src in self.survey.srcList:
            u_src = f[src, self._solutionType].copy()
            for rx in src.rxList:
                # wrt f, need possibility wrt m
                if v is not None:
                    PTv = rx.evalDeriv(
                        src, self.mesh, f, v[src, rx], adjoint=True
                    )
                else:
                    # This is for forming full sensitivity matrix
                    PTv = rx.getP(self.mesh, rx.projGLoc(f)).toarray().T
                df_duTFun = getattr(f, '_{0!s}Deriv'.format(rx.projField),
                                    None)
                df_duT, df_dmT = df_duTFun(src, None, PTv, adjoint=True)

                ATinvdf_duT = self.Ainv * df_duT

                dA_dmT = self.getADeriv(u_src, ATinvdf_duT, adjoint=True)
                dRHS_dmT = self.getRHSDeriv(src, ATinvdf_duT, adjoint=True)
                du_dmT = -dA_dmT + dRHS_dmT
                if v is not None:
                    Jtv += (df_dmT + du_dmT).astype(float)
                else:
                    iend = istrt + rx.nD
                    if rx.nD == 1:
                        Jtv[:, istrt] = (df_dmT + du_dmT)
                    else:
                        Jtv[:, istrt:iend] = (df_dmT + du_dmT)
                    istrt += rx.nD

        if v is not None:
            return Utils.mkvc(Jtv)
        else:
            # return np.hstack(Jtv)
            return Jtv

    def getSourceTerm(self):
        """
        Evaluates the sources, and puts them in matrix form
        :rtype: tuple
        :return: q (nC or nN, nSrc)
        """

        Srcs = self.survey.srcList

        if self._formulation == 'EB':
            n = self.mesh.nN
            # return NotImplementedError

        elif self._formulation == 'HJ':
            n = self.mesh.nC

        q = np.zeros((n, len(Srcs)))

        for i, src in enumerate(Srcs):
            q[:, i] = src.eval(self)
        return q

    @property
    def deleteTheseOnModelUpdate(self):
        toDelete = super(BaseDCProblem, self).deleteTheseOnModelUpdate
        if self._Jmatrix is not None:
            toDelete += ['_Jmatrix']
        return toDelete


class Problem3D_CC(BaseDCProblem):
    """
    3D cell centered DC problem
    """

    _solutionType = 'phiSolution'
    _formulation = 'HJ'  # CC potentials means J is on faces
    fieldsPair = Fields_CC
    bc_type = 'Neumann'

    def __init__(self, mesh, **kwargs):

        BaseDCProblem.__init__(self, mesh, **kwargs)
        self.setBC()

    def getA(self):
        """
        Make the A matrix for the cell centered DC resistivity problem
        A = D MfRhoI G
        """

        D = self.Div
        G = self.Grad
        MfRhoI = self.MfRhoI
        A = D * MfRhoI * G

        if(self.bc_type == 'Neumann'):
            Vol = self.mesh.vol
            if self.verbose:
                print('Perturbing first row of A to remove nullspace for Neumann BC.')

            # Handling Null space of A
            I, J, V = sp.sparse.find(A[0, :])
            for jj in J:
                A[0, jj] = 0.
            A[0, 0] = 1.

        # I think we should deprecate this for DC problem.
        # if self._makeASymmetric is True:
        #     return V.T * A
        return A

    def getADeriv(self, u, v, adjoint=False):

        D = self.Div
        G = self.Grad
        MfRhoIDeriv = self.MfRhoIDeriv

        if adjoint:
            return MfRhoIDeriv(G * u, D.T * v, adjoint)

        return D * (MfRhoIDeriv(G * u, v, adjoint))

    def getRHS(self):
        """
        RHS for the DC problem
        q
        """

        RHS = self.getSourceTerm()

        return RHS

    def getRHSDeriv(self, src, v, adjoint=False):
        """
        Derivative of the right hand side with respect to the model
        """
        # TODO: add qDeriv for RHS depending on m
        # qDeriv = src.evalDeriv(self, adjoint=adjoint)
        # return qDeriv
        return Zero()

    def setBC(self):
        if self.bc_type == 'Dirichlet':
            print('Homogeneous Dirichlet is the natural BC for this CC discretization.')
            self.Div = Utils.sdiag(self.mesh.vol) * self.mesh.faceDiv
            self.Grad = self.Div.T

        else:
            if self.mesh._meshType == "TREE" and self.bc_type == 'Neumann':
                raise NotImplementedError()

            if self.mesh.dim == 3:
                fxm, fxp, fym, fyp, fzm, fzp = self.mesh.faceBoundaryInd
                gBFxm = self.mesh.gridFx[fxm, :]
                gBFxp = self.mesh.gridFx[fxp, :]
                gBFym = self.mesh.gridFy[fym, :]
                gBFyp = self.mesh.gridFy[fyp, :]
                gBFzm = self.mesh.gridFz[fzm, :]
                gBFzp = self.mesh.gridFz[fzp, :]

                # Setup Mixed B.C (alpha, beta, gamma)
                temp_xm = np.ones_like(gBFxm[:, 0])
                temp_xp = np.ones_like(gBFxp[:, 0])
                temp_ym = np.ones_like(gBFym[:, 1])
                temp_yp = np.ones_like(gBFyp[:, 1])
                temp_zm = np.ones_like(gBFzm[:, 2])
                temp_zp = np.ones_like(gBFzp[:, 2])

                if(self.bc_type == 'Neumann'):
                    if self.verbose:
                        print('Setting BC to Neumann.')
                    alpha_xm, alpha_xp = temp_xm*0., temp_xp*0.
                    alpha_ym, alpha_yp = temp_ym*0., temp_yp*0.
                    alpha_zm, alpha_zp = temp_zm*0., temp_zp*0.

                    beta_xm, beta_xp = temp_xm, temp_xp
                    beta_ym, beta_yp = temp_ym, temp_yp
                    beta_zm, beta_zp = temp_zm, temp_zp

                    gamma_xm, gamma_xp = temp_xm*0., temp_xp*0.
                    gamma_ym, gamma_yp = temp_ym*0., temp_yp*0.
                    gamma_zm, gamma_zp = temp_zm*0., temp_zp*0.

                elif(self.bc_type == 'Dirichlet'):
                    if self.verbose:
                        print('Setting BC to Dirichlet.')
                    alpha_xm, alpha_xp = temp_xm, temp_xp
                    alpha_ym, alpha_yp = temp_ym, temp_yp
                    alpha_zm, alpha_zp = temp_zm, temp_zp

                    beta_xm, beta_xp = temp_xm*0, temp_xp*0
                    beta_ym, beta_yp = temp_ym*0, temp_yp*0
                    beta_zm, beta_zp = temp_zm*0, temp_zp*0

                    gamma_xm, gamma_xp = temp_xm*0., temp_xp*0.
                    gamma_ym, gamma_yp = temp_ym*0., temp_yp*0.
                    gamma_zm, gamma_zp = temp_zm*0., temp_zp*0.

                elif(self.bc_type == 'Mixed'):
                    # Ztop: Neumann
                    # Others: Mixed: alpha * phi + d phi dn = 0
                    # where alpha = 1 / r  * dr/dn
                    # (Dey and Morrison, 1979)

                    # This assumes that the source is located at
                    # (x_center, y_center_y, ztop)
                    # TODO: Implement Zhang et al. (1995)

                    xs = np.median(self.mesh.vectorCCx)
                    ys = np.median(self.mesh.vectorCCy)
                    zs = self.mesh.vectorCCz[-1]

                    def r_boundary(x, y, z):
                        return 1./np.sqrt(
                            (x - xs)**2 + (y - ys)**2 + (z - zs)**2
                        )
                    rxm = r_boundary(gBFxm[:, 0], gBFxm[:, 1], gBFxm[:, 2])
                    rxp = r_boundary(gBFxp[:, 0], gBFxp[:, 1], gBFxp[:, 2])
                    rym = r_boundary(gBFym[:, 0], gBFym[:, 1], gBFym[:, 2])
                    ryp = r_boundary(gBFyp[:, 0], gBFyp[:, 1], gBFyp[:, 2])
                    rzm = r_boundary(gBFzm[:, 0], gBFzm[:, 1], gBFzm[:, 2])

                    alpha_xm = (gBFxm[:, 0]-xs)/rxm**2
                    alpha_xp = (gBFxp[:, 0]-xs)/rxp**2
                    alpha_ym = (gBFym[:, 1]-ys)/rym**2
                    alpha_yp = (gBFyp[:, 1]-ys)/ryp**2
                    alpha_zm = (gBFzm[:, 2]-zs)/rzm**2
                    alpha_zp = temp_zp.copy() * 0.

                    beta_xm, beta_xp = temp_xm*1, temp_xp*1
                    beta_ym, beta_yp = temp_ym*1, temp_yp*1
                    beta_zm, beta_zp = temp_zm*1, temp_zp*1

                    gamma_xm, gamma_xp = temp_xm*0., temp_xp*0.
                    gamma_ym, gamma_yp = temp_ym*0., temp_yp*0.
                    gamma_zm, gamma_zp = temp_zm*0., temp_zp*0.

                alpha = [
                    alpha_xm, alpha_xp,
                    alpha_ym, alpha_yp,
                    alpha_zm, alpha_zp
                ]
                beta = [beta_xm, beta_xp, beta_ym, beta_yp, beta_zm, beta_zp]
                gamma = [gamma_xm, gamma_xp, gamma_ym, gamma_yp, gamma_zm,
                         gamma_zp]

            elif self.mesh.dim == 2:

                fxm, fxp, fym, fyp = self.mesh.faceBoundaryInd
                gBFxm = self.mesh.gridFx[fxm, :]
                gBFxp = self.mesh.gridFx[fxp, :]
                gBFym = self.mesh.gridFy[fym, :]
                gBFyp = self.mesh.gridFy[fyp, :]

                # Setup Mixed B.C (alpha, beta, gamma)
                temp_xm = np.ones_like(gBFxm[:, 0])
                temp_xp = np.ones_like(gBFxp[:, 0])
                temp_ym = np.ones_like(gBFym[:, 1])
                temp_yp = np.ones_like(gBFyp[:, 1])

                alpha_xm, alpha_xp = temp_xm*0., temp_xp*0.
                alpha_ym, alpha_yp = temp_ym*0., temp_yp*0.

                beta_xm, beta_xp = temp_xm, temp_xp
                beta_ym, beta_yp = temp_ym, temp_yp

                gamma_xm, gamma_xp = temp_xm*0., temp_xp*0.
                gamma_ym, gamma_yp = temp_ym*0., temp_yp*0.

                alpha = [alpha_xm, alpha_xp, alpha_ym, alpha_yp]
                beta = [beta_xm, beta_xp, beta_ym, beta_yp]
                gamma = [gamma_xm, gamma_xp, gamma_ym, gamma_yp]

            x_BC, y_BC = getxBCyBC_CC(self.mesh, alpha, beta, gamma)
            V = self.Vol
            self.Div = V * self.mesh.faceDiv
            P_BC, B = self.mesh.getBCProjWF_simple()
            M = B*self.mesh.aveCC2F
            self.Grad = self.Div.T - P_BC*Utils.sdiag(y_BC)*M


class Problem3D_N(BaseDCProblem):
    """
    3D nodal DC problem
    """

    _solutionType = 'phiSolution'
    _formulation = 'EB'  # N potentials means B is on faces
    fieldsPair = Fields_N

    def __init__(self, mesh, **kwargs):
        BaseDCProblem.__init__(self, mesh, **kwargs)
        # Not sure why I need to do this
        # To evaluate mesh.aveE2CC, this is required....
        if mesh._meshType == "TREE":
            mesh.nodalGrad

    def getA(self):
        """
        Make the A matrix for the cell centered DC resistivity problem
        A = G.T MeSigma G
        """

        MeSigma = self.MeSigma
        Grad = self.mesh.nodalGrad
        A = Grad.T * MeSigma * Grad

        # Handling Null space of A
        I, J, V = sp.sparse.find(A[0, :])
        for jj in J:
            A[0, jj] = 0.
        A[0, 0] = 1.

        return A

    def getADeriv(self, u, v, adjoint=False):
        """
        Product of the derivative of our system matrix with respect to the
        model and a vector
        """
        Grad = self.mesh.nodalGrad
        if not adjoint:
            return Grad.T*self.MeSigmaDeriv(Grad*u, v, adjoint)
        elif adjoint:
            return self.MeSigmaDeriv(Grad*u, Grad*v, adjoint)

    def getRHS(self):
        """
        RHS for the DC problem
        q
        """

        RHS = self.getSourceTerm()
        return RHS

    def getRHSDeriv(self, src, v, adjoint=False):
        """
        Derivative of the right hand side with respect to the model
        """
        # TODO: add qDeriv for RHS depending on m
        # qDeriv = src.evalDeriv(self, adjoint=adjoint)
        # return qDeriv
        return Zero()
